/*****************************************
 *  Copyright  2001-2007
 *  Sigma Designs, Inc. All Rights Reserved
 *  Proprietary and Confidential
 ******************************************/

#include <linux/init.h>
#include <linux/slab.h>
#include <linux/irq.h>
#include <linux/poll.h>
#include <linux/delay.h>

#if !defined(CONFIG_TANGOX)
#error "Unsupport architecture (TANGOX only)."
#elif defined(CONFIG_TANGO2_SMP863X) && (EM86XX_REVISION < 4)
#error "Unsupport Tango2 (ES4 or above) or Tango3 chip."
#endif

#include <asm/io.h>

#ifdef CONFIG_TANGO2
#include <asm/tango2/tango2_gbus.h>
#include <asm/tango2/tango2api.h>
#include <asm/tango2/scard.h>
#define NUM_SCARD          1
#elif defined(CONFIG_TANGO3)
#include <asm/tango3/tango3_gbus.h>
#include <asm/tango3/tango3api.h>
#include <asm/tango3/scard.h>
#define NUM_SCARD          2
#else
#error "Unsupport architecture."
#endif


//#define CLASS_B // 3V
//#define PARSE_ATR
#define READ_TIMEOUT

#define RW_TIMEOUT 600 // x5mS = 3SEC

//#define SCARD_DEBUG /* Uncomment this for debug message */
#define SCARD_ERRMSG /* Uncomment this for error message */
#define SCARD_MSG /* Uncomment this for normal message */
#define SCARD_DEBUG_X1 /* Uncomment this for debug xx message */

#ifdef SCARD_DEBUG
#define DBG_PRINT	if (debug) printk
#else
static void inline DBG_PRINT(const char *x, ...) { ; }
#endif /* SCARD_DEBUG */

#ifdef SCARD_ERRMSG
#define ERR_PRINT	printk
#else
static void inline ERR_PRINT(const char *x, ...) { ; }
#endif /* SCARD_ERRMSG */

#ifdef SCARD_MSG
#define MSG_PRINT	printk
#else
static void inline MSG_PRINT(const char *x, ...) { ; }
#endif /* SCARD_ERRMSG */

#ifdef SCARD_DEBUG_X1
#define DBG_X_PRINT	printk
#else
static void inline DBG_X_PRINT(const char *x, ...) { ; }
#endif /* SCARD_DEBUG */


#define SCARD_TX_BYTE_REG(x)	((x) + 0x0000)
#define SCARD_TX_WORD_REG(x)	((x) + 0x0004)
#define SCARD_TX_DWORD_REG(x)	((x) + 0x0008)
#define SCARD_RX_BYTE_REG(x)	((x) + 0x0010)
#define SCARD_RX_WORD_REG(x)	((x) + 0x0014)
#define SCARD_RX_DWORD_REG(x)	((x) + 0x0018)
#define SCARD_CLK_HIGH_VAL(x)	((x) + 0x0020)
#define SCARD_CLK_LOW_VAL(x)	((x) + 0x0024)
#define SCARD_TIMEOUT_LOAD(x)	((x) + 0x0028)
#define SCARD_PARAM_REG(x)	((x) + 0x002c)
#define SCARD_EGT_ETU_REG(x)	((x) + 0x0030)
#define SCARD_SOFT_OUT_REG(x)	((x) + 0x0034)
#define SCARD_SCARD_CTRL_REG(x)	((x) + 0x0038)
#define SCARD_STATE_REG(x)	((x) + 0x003c)
#define SCARD_INT_REG(x)	((x) + 0x0040)
#define SCARD_INTEN_REG(x)	((x) + 0x0044)
#define SCARD_ALT_ETU_REG(x)	((x) + 0x0048)
#define SCARD_STATUS_REG(x)	((x) + 0x004c)
#define SCARD_CRC_REG(x)	((x) + 0x0050)
#define SCARD_CRC_INIT_REG(x)	((x) + 0x0054)

#ifdef CONFIG_TANGO3
#define SCARD_GPIO_OE_REG(x)       ((x) + 0x0058)
#define SCARD_GPIO_DATA_REG(x)     ((x) + 0x005c)
#define SCARD_GPIO_MODE_REG(x)     ((x) + 0x0060)
#define SCARD_RX_TO_TX_IDLE_REG(x) ((x) + 0x0064)
#endif

#define RD_BUF_SIZE		256	/* Default read buffer size */
#define WR_BUF_SIZE		256	/* Default write buffer size */
#define ANS_STR_SIZE		512	/* Max. size of ans-to-reset string */
#define DEF_FREQUENCY           3600000 // 3571200 //6000000 //2000000 /* Default operating frequency 2MHz */
#define TIMEOUT                 ((((2 * HZ) / 1000) == 0) ? 1 : ((2 * HZ) / 1000))

/* The scard control pins: can be board dependent */
#define SCARD_OFF_PIN           0  /* SCARD_CTL[0] */
#define SCARD_5V_PIN            1  /* SCARD_CTL[1] */
#define SCARD_CMD_PIN           2  /* SCARD_CTL[2] */

#define SCARD2_OFF_PIN          0  /* SCARD_CTL[0] */
#define SCARD2_5V_PIN           1  /* SCARD_CTL[1] */
#define SCARD2_CMD_PIN          2  /* SCARD_CTL[2] */

/* The major device number and name */
#define SCARD_DEV_MAJOR		182	
#define SCARD_DEV_NAME		"scard"

#define RST_WAIT		64    //<-2k90106 100 //64	/* Wait time to drive RSTIN low */
					/* Min. 50, Max. 130 */
#define RST_WIDTH		1000  //<-2k90106 100 //1	/* Keep RSTIN for this long */

#define TX_FIFO_SIZE            16 ///0x0f

#define DRIVER_VERSION        "1.6.5" ///// org "1.29"
#define DRIVER_DATE         	"2009-07-22"

MODULE_DESCRIPTION("EM86xx/SMP86xx smart card driver\n");
MODULE_AUTHOR("TANGO15/TANGO2/TANGO3 standalone team");
MODULE_LICENSE("GPL");
MODULE_VERSION(DRIVER_VERSION);

MODULE_PARM(major_num, "i");
MODULE_PARM(rd_buffer_size, "i");
MODULE_PARM(wr_buffer_size, "i");
MODULE_PARM(frequency, "i");
MODULE_PARM(reverse_msb, "i");
MODULE_PARM(reverse_pol, "i");
MODULE_PARM(use_3v, "i");
#ifdef SCARD_DEBUG
MODULE_PARM(debug, "i");
#endif
static int rd_buffer_size = RD_BUF_SIZE;
static int wr_buffer_size = WR_BUF_SIZE;
static int frequency = DEF_FREQUENCY;
static int reverse_msb = 0; /* By default, don't reverse MSB */
static int reverse_pol = 0; /* By default, don't reverse polarity */

#ifdef CLASS_B
static int use_3v = 1; /* By default, start with 3V */
#else // Only Clsss A
static int use_3v = 0;///<-Only 5V
#endif

#ifdef SCARD_DEBUG
static int debug = 1;
#endif

#ifdef READ_TIMEOUT
static int Read_act = 0, Read_act_count = 0;
#endif

static int Last_rev_msb = 0, Last_rev_pol = 0; 

static unsigned char Tx_Mode = 0; // 0 is T=0, 1 is T=1, 2 is T=14

static unsigned char Reset_mode_count = 0, Reset_Fail_count = 0, IRD_ATR_X_mode = 0;

static unsigned char RW_Abort = 0;

static unsigned int RW_MORE_DELAY = 0;

static unsigned char T0_ACK_Master = 0;
static unsigned char T0_Last_Write_Size = 0;

#ifdef CONFIG_TANGO3
#define MAX_SC	2
#else
#define MAX_SC	1
#endif

static unsigned char header_read = 0, ISO_Non_T0_Act = 0, write_ready = 0;

static unsigned char ATRXBuf[5]={0,}, ATRXBuf_CNT = 0; //2k90120

static unsigned char Remove_Act = 0, Remove_Act_wait_count = 0;

#ifdef SCARD_DEBUG_X1
unsigned char ISR_DBG_Count = 0;
#endif

extern unsigned long tangox_get_sysclock(void);

/* The state for smart card */
enum SCARD_STATE {
	UNPWR_IDLE = 0,
	ACTIVATE = 1,
	COLD_RESET = 2,
	PWR_IDLE = 3,
	WARM_RESET = 4,
	CLOCK_STOP = 5,
	DEACTIVATE = 6,
	UNKNOWN = 7,
};

/* If given state is persistent state? */
#define PERSISTENT_STATE(x)	(((x) == UNPWR_IDLE) || ((x) == CLOCK_STOP) || ((x) == PWR_IDLE))

/* Private data structure */
struct scard_private {
	unsigned char *rd_buffer;	/* Circular buffer for read */
	unsigned char *wr_buffer;	/* Circular buffer for write */
	unsigned rd_p_idx;		/* Index of producer */
	unsigned rd_c_idx; 		/* Index of consumer */
	unsigned wr_p_idx;		/* Index of producer */
	unsigned wr_c_idx; 		/* Index of consumer */
	unsigned ref_cnt;		/* Reference count */
	spinlock_t lock;		/* Spin lock */
	unsigned char b_mode;		/* Blocking mode or not */
	unsigned long last_rd_jiffies;	/* Timestamp for last reception */
	unsigned long last_wr_jiffies;	/* Timestamp for last reception */
	atomic_t card_existed;		/* Is smart card existed? */
	atomic_t state;			/* State? UNPWR_IDLE or PWR_IDLE */
	atomic_t normal_mode;		/* Normal operation mode? */
	int rev_msb;			/* Reverse MSB? */
	int rev_pol;			/* Reverse polarity? */
	int volt3;			/* 3V card? */
	int clockstop;			/* clockstop supported */
	int reset_cnt;                  /* reset counter */
	int stop_thread;                
	unsigned long base_addr;        /* base address for scard controller */
	unsigned long write_cnt;
	unsigned long read_cnt;
	int scard_5v_pin;
	int scard_cmd_pin;
	int scard_off_pin;
	int scard_irq;
	char devname[32];
	int idx;                        /* controller index */

	/* For answer to reset */
	atomic_t answer2reset;
	unsigned char ans_string[ANS_STR_SIZE];
	int ans_idx;

	/* Wait queue, may be used if block mode is on */
	wait_queue_head_t scard_rd_wq;
};

/* Static information for each scard controller */
struct scard_info {
	unsigned long base_addr;
	int scard_5v_pin;
	int scard_cmd_pin;
	int scard_off_pin;
	int scard_irq;
} scard_ctrlers[] = {
	{ (REG_BASE_cpu_block + 0xc300), SCARD_5V_PIN, SCARD_CMD_PIN, SCARD_OFF_PIN, IRQ_CONTROLLER_IRQ_BASE + LOG2_CPU_SMARTCARD_INT },
#ifdef CONFIG_TANGO3
	{ (REG_BASE_cpu_block + 0xc380), SCARD2_5V_PIN, SCARD2_CMD_PIN, SCARD2_OFF_PIN, IRQ_CONTROLLER_IRQ_BASE + LOG2_CPU_SMARTCARD2_INT },
#endif
};

/* Some prototypes */
static int scard_open(struct inode *, struct file *);
static int scard_release(struct inode *, struct file *);
static int scard_read(struct file *, char *, size_t, loff_t *);
static int scard_write(struct file *, const char *, size_t, loff_t *);
static int scard_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
static unsigned int scard_poll(struct file *, struct poll_table_struct *);

#if 0 // 2k80706
static int scard_ext_ioctl(struct inode *inode, struct file *fptr, unsigned char Mode, char *buffer);
#endif

/* Global data */
static struct scard_private scard_priv[NUM_SCARD];
static char *scard_devname = SCARD_DEV_NAME;
static int major_num = SCARD_DEV_MAJOR;

int tangox_get_scard_info(int *pin_5v, int *pin_cmd, int *pin_off);

static int xenv_scard_5v_pin, xenv_scard_off_pin, xenv_scard_cmd_pin;

/* From smartcard spec. */
static const unsigned fi_tab[16] = { 372, 372, 558, 744, 1116, 1488, 1860, 0, 0, 512, 768, 1024, 1536, 2048, 0, 0 };
static const unsigned di_tab[16] = { 0, 1, 2, 4, 8, 16, 32, 0, 12, 20, 0, 0, 0, 0, 0, 0 };

/* Keep track of system frequency */
static unsigned long sys_freq = 0;

static struct file_operations scard_fops = {
	open: scard_open,
	read: scard_read,
	write: scard_write,
	ioctl: scard_ioctl,
	poll: scard_poll,
	release: scard_release,
	owner: THIS_MODULE,
};

/* Prototype */
static void scard_reset(struct scard_private *priv, int cold);

static inline unsigned int scard_ctrlpin_read(int idx, unsigned int pin)
{
	unsigned int val;
#ifdef CONFIG_TANGO2
	val = (gbus_read_uint32(pGBus, REG_BASE_cpu_block + CPU_uart0_gpio_data) >> 16) & 0x7;
#elif defined(CONFIG_TANGO3)
	val = (gbus_read_uint32(pGBus, SCARD_GPIO_DATA_REG(scard_priv[idx].base_addr)) >> 4) & 0x7;
#endif
	return((val >> pin) & 0x1);
}

static inline void scard_ctrlpin_write(int idx, int pin, unsigned int val)
{
	unsigned int v;
#ifdef CONFIG_TANGO2
	v = ((gbus_read_uint32(pGBus, REG_BASE_cpu_block + CPU_uart0_gpio_data) >> 16) & 0x7) & ~(1 << pin);
	v |= (((val != 0) ? 1 : 0) << pin);
	gbus_write_uint32(pGBus, REG_BASE_cpu_block + CPU_uart0_gpio_data, v << 16);
#elif defined(CONFIG_TANGO3)
	v = (((val != 0) ? 1 : 0) << (pin + 4)) | (1 << (pin + 12));
	gbus_write_uint32(pGBus, SCARD_GPIO_DATA_REG(scard_priv[idx].base_addr), v);
#endif
}

static inline void scard_ctrlpin_init(int idx)
{
	/* Set SCARD_CTL[0..2] to 2 outputs for 5V, CMD, and 1 input for OFF */
#ifdef CONFIG_TANGO2
	gbus_write_uint32(pGBus, REG_BASE_cpu_block + CPU_uart0_gpio_dir,
			((1 << scard_priv[idx].scard_cmd_pin) | (1 << scard_priv[idx].scard_5v_pin)) << 16);
#elif defined(CONFIG_TANGO3)
	unsigned int v = ((1 << scard_priv[idx].scard_cmd_pin) | (1 << scard_priv[idx].scard_5v_pin)) << 4;
	v |= 0x7000; /* With OE mask set */
	gbus_write_uint32(pGBus, SCARD_GPIO_MODE_REG(scard_priv[idx].base_addr), 0x7f00); /* Set smard card mode */
	gbus_write_uint32(pGBus, SCARD_GPIO_OE_REG(scard_priv[idx].base_addr), v); /* Set outputs */
#endif
}

/* Read data from device */
static void scard_read_device(struct scard_private *priv)
{
	unsigned char data = 0;
	unsigned rd_pidx;
	unsigned int fifo_size;
	unsigned int i;

	spin_lock(&priv->lock);
	if ((fifo_size = ((gbus_read_uint32(pGBus, SCARD_STATE_REG(priv->base_addr)) >> 20) & 0x1f)) == 0) {
		DBG_PRINT("%s: RX FIFO empty\n", priv->devname);
		goto out;
	}
	
	DBG_PRINT("%s: scard_read_device fifo_size = %d\n", priv->devname, fifo_size);
	/* Read from FIFO into driver's buffer */
	for (i = 0; i < fifo_size; i++) {
		data = gbus_read_uint32(pGBus, SCARD_RX_BYTE_REG(priv->base_addr));
		priv->read_cnt++;

		DBG_PRINT("%s: got data 0x%x (cnt: %ld)\n", priv->devname, data, priv->read_cnt);
		rd_pidx = priv->rd_p_idx;	/* Save the old index before proceeding */

		/* Save it to buffer */
		if (((priv->rd_p_idx + 1) % rd_buffer_size) == priv->rd_c_idx) {
			ERR_PRINT("%s: read buffer full\n", priv->devname);
			break;
		}

		priv->rd_buffer[priv->rd_p_idx] = data;
		priv->rd_p_idx = (priv->rd_p_idx + 1) % rd_buffer_size;

		/* Buffer was empty and block mode is on, wake up the reader */
		if ((priv->b_mode != 0) && (priv->rd_c_idx == rd_pidx)) 
			wake_up_interruptible(&(priv->scard_rd_wq));
	}
	priv->last_rd_jiffies = jiffies;

out:
	spin_unlock(&priv->lock);
}

/* Wrtie data to device */
static void scard_write_device(struct scard_private *priv)
{
	unsigned char data;
	unsigned int fifo_size;
	unsigned long flags;
	unsigned int i;

	spin_lock_irqsave(&priv->lock, flags);

	if (priv->wr_c_idx == priv->wr_p_idx) {
		goto out; /* Empty buffer */
	} else if ((fifo_size = ((gbus_read_uint32(pGBus, SCARD_STATE_REG(priv->base_addr)) >> 12) & 0x1f)) == TX_FIFO_SIZE) {
		DBG_PRINT("%s: TX FIFO full (%d)\n", priv->devname, fifo_size);
		goto out;
	} else if (fifo_size > TX_FIFO_SIZE) {
		ERR_PRINT("%s: TX FIFO overflow (%d)\n", priv->devname, fifo_size);
		goto out;
	}
	
	
#if 1 // 2k80717
	//////DBG_X_PRINT("W:FIFO_RST\n");
	/* Reset the RX/TX FIFOs */
	gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
				gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) | ((1 << 24) | (1 << 25)));
	gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
				gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) & ~((1 << 24) | (1 << 25)));
#endif

	//priv->write_cnt=0; // 2k80704
	
	for (i = 0; i < (TX_FIFO_SIZE - fifo_size); i++) {
		if (priv->wr_c_idx == priv->wr_p_idx) 
			break; /* Empty buffer */

		data = priv->wr_buffer[priv->wr_c_idx];
		priv->wr_c_idx = (priv->wr_c_idx + 1) % wr_buffer_size;
		gbus_write_uint32(pGBus, SCARD_TX_BYTE_REG(priv->base_addr), data);
		priv->write_cnt++;

		/*DBG_PRINT("%s: written data 0x%x (fifo: %d, cnt: %ld)\n", priv->devname, data, fifo_size, priv->write_cnt);*/
	}

/*#ifdef SCARD_DEBUG
	if (fifo_size >= TX_FIFO_SIZE)
		DBG_PRINT("%s: TX FIFO full %d\n", priv->devname, fifo_size);
#endif*/
	
	priv->last_wr_jiffies = jiffies;

out:
	spin_unlock_irqrestore(&priv->lock, flags);

	write_ready = 0;
}

/* ISR for Smart card device */
static irqreturn_t scard_isr(int irq, void *dev_id, struct pt_regs *regs)
{
	struct scard_private *priv = (struct scard_private *)dev_id;
	unsigned long status = 0;
	unsigned char rx_out_count=0, rx_max_count=0;

#ifdef SCARD_DEBUG_X1
	ISR_DBG_Count++;
#endif

	if ((status = gbus_read_uint32(pGBus, SCARD_INT_REG(priv->base_addr))) == 0)
		return IRQ_NONE;

	gbus_write_uint32(pGBus, SCARD_INT_REG(priv->base_addr), status);
	
	if (atomic_read(&(priv->normal_mode)) != 0) {
		if (status & (1 << 2)) { 
			/* For RX interrupts */
			scard_read_device(priv);
		}
		if (status & (1 << 0)) {
			/* For TX interrupts */
			scard_write_device(priv);
		}
		if (status & ((1 << 1) | (1 << 3) | (1 << 4) | (1 << 5))) {
			#if 0
			ERR_PRINT("%s: error interrupt (status: 0x%08lx)\n", priv->devname, status);
			#endif
			/* TODO: does WARM_RESET, reset/un-reset FIFO ?? */
		}
	} else {
		/* Initializing connection, handle "answer to reset" */
		atomic_set(&(priv->answer2reset), 1);
		if (status & (1 << 0)) {
			/* For TX interrupts */
			DBG_PRINT("TX ..\n");
		}
		if (status & (1 << 2)) { 
			unsigned long data;

#if 0//def SCARD_DEBUG_X1
			DBG_X_PRINT(" RX..\n");
#endif

#if 0 // org
			rx_out_count=rx_max_count=0;
			while (((gbus_read_uint32(pGBus, SCARD_STATE_REG(priv->base_addr)) >> 20) & 0x1f) != 0) {
				/* For RX interrupts */
				data = gbus_read_uint32(pGBus, SCARD_RX_BYTE_REG(priv->base_addr)); 

#if 0//def SCARD_DEBUG_X1
				DBG_X_PRINT("0x%02lX  ",data);
				udelay(1000);
#else
				//if((Tx_Mode == 2)||(Reset_Fail_count != 0)) // T=14 or No ATR
				{
					#if 0
					udelay(2000); // ?? 2k80703
					///#else // 2k90109
					udelay(1000);
					udelay(1000);
					udelay(1000);
//					udelay(1000);
//					udelay(1000);
					#endif
				}
#endif

				if (priv->ans_idx >= ANS_STR_SIZE) {
					gbus_write_uint32(pGBus, SCARD_INTEN_REG(priv->base_addr), 0); /* disable interrupt */
					break;
				}
				priv->ans_string[priv->ans_idx++] = data & 0xff;
			}
#else // 2k90127
			/*unsigned int fifo_size;*/
			unsigned char Empty_count = 2;

			rx_max_count = rx_out_count = 0;

			if(IRD_ATR_X_mode == 1)
			{
				Empty_count = 30;
			}
			else
			{
				Empty_count = 2;
			}

			while(1)
			{
				if(((gbus_read_uint32(pGBus, SCARD_STATE_REG(priv->base_addr)) >> 20) & 0x1f) == 0)
				/*if ((fifo_size = ((gbus_read_uint32(pGBus, SCARD_STATE_REG(priv->base_addr)) >> 20) & 0x1f)) == 0)*/
				{
					if(rx_out_count++ > Empty_count)
					{
						///DBG_X_PRINT(">>>>>>> rx_out_count limit over=%d\n", rx_out_count);
						break;
					}
					else
					{
						if(IRD_ATR_X_mode == 0)
						{
							udelay(500);
						}
						else
						{
							udelay(1000);
							udelay(1000);
							udelay(1000);	
							udelay(1000);
							udelay(1000);
						}
						continue;
					}
				}

				/*if(fifo_size > 1) DBG_X_PRINT("--------->> fifo_size = %d\n",fifo_size); */

				if(rx_max_count++ > 250)
				{
					DBG_X_PRINT(">>>>>>> Warning -- rx_max_count limit over=%d\n", rx_max_count);
					break;
				}

				/* For RX interrupts */
				data = gbus_read_uint32(pGBus, SCARD_RX_BYTE_REG(priv->base_addr)); 

#if 0//def SCARD_DEBUG_X1
				DBG_X_PRINT("0x%02lX  ",data);
				udelay(1000);
#else
				/// Disable 2k90202 if((Tx_Mode == 2)||(Reset_Fail_count != 0)) // T=14 or No ATR
				{
					udelay(1000);
					udelay(1000);
					udelay(1000);
				}
#endif

				if (priv->ans_idx >= ANS_STR_SIZE) {
					gbus_write_uint32(pGBus, SCARD_INTEN_REG(priv->base_addr), 0); /* disable interrupt */
					break;
				}
				priv->ans_string[priv->ans_idx++] = data & 0xff;
			}
#endif

#if 0//def SCARD_DEBUG_X1
			DBG_X_PRINT(" End..\n");
#endif
		}
		if (status & ((1 << 1) | (1 << 3) | (1 << 4) | (1 << 5))) {
			gbus_write_uint32(pGBus, SCARD_INTEN_REG(priv->base_addr), 0); /* disable interrupt */
			DBG_PRINT("%s: error interrupt (status: 0x%08lx)\n", priv->devname, status);
			DBG_PRINT("%s: SCARD_CTRL : 0x%08lx\n", priv->devname, gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)));
		}
	}

	return IRQ_HANDLED;
}

/* Check if scard existed */
static int is_card_present(struct scard_private *priv)
{
	/* Low: not present, high otherwise */
	return((scard_ctrlpin_read(priv->idx, priv->scard_off_pin) == 0) ? 0 : 1);
}

#ifdef PARSE_ATR
/* Parse and handle "answer to reset" string */
static int parse_ans2reset(struct scard_private *priv, int v3, const unsigned char str[], const int len)
{
	int idx = 2, i = 1;
	int is_ta = 0, is_tb = 0, is_tc = 0, is_td = 0, is_tck = 0, k = 0;
	unsigned char ta = 0xff, tb = 0xff, tc = 0xff, td = 0xff, tx = 0xff;
	unsigned char tck = 0xff;
	unsigned char t_array[16] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
	
	priv->volt3 = v3;
	switch(str[0]) {
		case 0x3b: /* Correct config detected */
		case 0x3f:
			DBG_PRINT("Correct ts detected 0x%x\n", str[0]);
			break;
		case 0xc4:
		case 0xc0:
			DBG_PRINT("Wrong ts detected 0x%x (need to reverse the polarity)\n", str[0]);
			priv->rev_pol = 1;
			goto reset_needed;
			break;
		case 0xdc:
		case 0xfc:
			DBG_PRINT("Wrong ts detected 0x%x (need to reverse the MSB)\n", str[0]);
			priv->rev_msb = 1;
			goto reset_needed;
			break;
		case 0x23:
		case 0x03:
			DBG_PRINT("Wrong ts detected 0x%x (need to reverse the polarity and MSB)\n", str[0]);
			priv->rev_msb = priv->rev_pol = 1;
			goto reset_needed;
			break;
		default:
			priv->reset_cnt++;
			if (priv->reset_cnt > 3)
				goto not_valid;
			else {
				switch(priv->reset_cnt) {
					case 1: priv->rev_pol = (priv->rev_pol != 0) ? 0 : 1;
						break;
					case 2: priv->rev_msb = (priv->rev_msb != 0) ? 0 : 1;
						break;
					case 3: priv->rev_pol = (priv->rev_pol != 0) ? 0 : 1;
						priv->rev_msb = (priv->rev_msb != 0) ? 0 : 1;
						break;
				}
				goto reset_needed;
			}
			break;
	}
	k = str[1] & 0xf;

	DBG_PRINT("Parsing answer string: t0=0x%x k=%d\n", str[1], k);

	for (tx = str[1]; idx < len; i++) {
		is_ta = ((tx & 0x10) != 0);	
		is_tb = ((tx & 0x20) != 0);
		is_tc = ((tx & 0x40) != 0);
		is_td = ((tx & 0x80) != 0);
		DBG_PRINT("*** Iteration %d (idx = %d)\n", i, idx);
#ifdef SCARD_DEBUG
               {
                       int j;
                       DBG_PRINT("Parsing: ");
                       for (j = idx; j < len; j++)
                               DBG_PRINT("0x%x ", str[j]);
                       DBG_PRINT("\n");
               }
#endif

		if (i == 1) {
			unsigned char fi = 0xff, di = 0xff, ii = 0xff, pi1 = 0xff, n = 0xff;
			if (is_ta != 0) {
				ta = str[idx++];
				fi = (ta & 0xf0) >> 4;
				di = (ta & 0xf);
				/* Resetting ETU value */
				DBG_PRINT("Reset ETU to %d\n", (fi_tab[fi] / di_tab[di]) - 1);
				gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), 
						(gbus_read_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr)) & 0xffff0000) | (((fi_tab[fi] / di_tab[di]) - 1) << 0));
				DBG_PRINT("ta=0x%x (fi=0x%x di=0x%x) ", ta, fi, di);
			}
			if (is_tb != 0) { 
				tb = str[idx++];
				ii = (tb & 0x60) >> 5;
				pi1 = (tb & 0x1f);
				DBG_PRINT("tb=0x%x (ii=0x%x pi1=0x%x) ", tb, ii, pi1);
			}
			if (is_tc != 0) {
				tc = str[idx++];
				n = tc;
				DBG_PRINT("tc=0x%x (n=0x%x) ", tc, n);
			}
			if (is_td != 0) {
				td = str[idx++];
				t_array[td & 0xf]++;
				tx = td;
				DBG_PRINT("td=0x%x", td);
			} else {
				t_array[0]++;
			}
			DBG_PRINT("\n");
			if (is_tc != 0) {
				if (n == 0xff) {
					if (t_array[0] != 0) {
						n = 3 - 1;
						DBG_PRINT("Reset EGT to %d\n", n);
						gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), (gbus_read_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr)) & 0xffff) | (n << 16));
					} else if (t_array[1] != 0) {
						n = 2 - 1;
						DBG_PRINT("Reset EGT to %d\n", n);
						gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), (gbus_read_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr)) & 0xffff) | (n << 16));
					}
				} else {
					if (t_array[15] == 0) {
						n = 2 + n - 1;
						DBG_PRINT("Reset EGT to %d\n", n);
						gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), (gbus_read_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr)) & 0xffff) | (n << 16));
					} else {
						n = 2 + n - 1;
						DBG_PRINT("Reset EGT to %d\n", n);
						gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), (gbus_read_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr)) & 0xffff) | (n << 16));
					}
				}
			}
		} else if (i == 2) {
			unsigned char pi2 = 0xff;
			if (is_ta != 0) {
				ta = str[idx++];
				DBG_PRINT("ta=0x%x ", ta);
			}
			if (is_tb != 0) { 
				tb = str[idx++];
				pi2 = tb;
				DBG_PRINT("tb=0x%x (pi2=0x%x) ", tb, pi2);
			}
			if (is_tc != 0) {
				tc = str[idx++];
				DBG_PRINT("tc=0x%x ", tc);
			}
			if (is_td != 0) {
				td = str[idx++];
				t_array[td & 0xf]++;
				tx = td;
				DBG_PRINT("td=0x%x", td);
			}
			DBG_PRINT("\n");
			if ((is_ta != 0) && ((ta & 0x80) != 0) && ((ta & 0x0f) == 1) && (t_array[1] != 0)) {
				DBG_PRINT("Disabling txerr_check, rxerr_assert.\n");
				gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
						gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) & ~((1 << 6) | (1 << 16)));
			}
		} else {
			unsigned char xi = 0xff, ui = 0xff;
			if (is_ta != 0) {
				ta = str[idx++];
				if (t_array[15] != 0) {
					xi = (ta & 0xc0) >> 6;
					ui = (ta & 0x3f);
					/* xi = 0 <-- no clock stop */
					/* xi = 1 <-- clock stop on low */
					/* xi = 2 <-- clock stop on high */
					/* xi = 3 <-- NA */
					if (xi == 1) { 
						gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
								gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) & ~(1 << 28));
						priv->clockstop = 1;
					} else if (xi == 2) {
						gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
								gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) | (1 << 28));
						priv->clockstop = 1;
					} else
						priv->clockstop = 0;
					DBG_PRINT("ta=0x%x (xi=0x%x ui=0x%x) ", ta, xi, ui);
				} else {
					DBG_PRINT("ta=0x%x ", ta);
				}
			}
			if (is_tb != 0) { 
				tb = str[idx++];
				DBG_PRINT("tb=0x%x ", tb);
			}
			if (is_tc != 0) {
				tc = str[idx++];
				DBG_PRINT("tc=0x%x ", tc);
			}
			if (is_td != 0) {
				td = str[idx++];
				t_array[td & 0xf]++;
				tx = td;
				DBG_PRINT("td=0x%x", td);
			}
			DBG_PRINT("\n");
		}

		if (is_td == 0) 
			break;
		else {
			is_ta = is_tb = is_tc = is_td = 0;
			ta = tb = tc = td = 0xff;
		}
	}		

	idx += k;

#ifdef SCARD_DEBUG
	DBG_PRINT("T = x is set for ");
	for (i = 0; i < 16; i++) {
		if (t_array[i] != 0)
			DBG_PRINT("%d ", i);
	}
	DBG_PRINT("\n");
#endif
	
	if (t_array[0] != 0) {
		for (i = 1; i < 16; i++) {
			if (t_array[i] != 0)
				break;
		}
		if (i < 16) {
			is_tck = 1;
			tck = str[idx++];
		}
	} else {
		is_tck = 1;
		tck = str[idx++];
	}

	if (idx != len) {
		/* Check to see if T=0 protocol is used */
		if ((t_array[0] != 0) && ((idx + 2) == len)) {
			if ((str[idx] != 0x90) || (str[idx + 1] != 0x00)) {
				DBG_PRINT("Length mismatched: %d vs. %d\n", idx, len);
				goto not_valid;
			}
		} else {
			DBG_PRINT("Length mismatched: %d vs. %d\n", idx, len);
			goto not_valid;
		}
	} 
	if (is_tck != 0) {
		for (k = 0, i = 1; i < len; i++)
			k ^= str[i];
		if (k != 0) {
			DBG_PRINT("TCK sum mismatched: 0x%x\n", k);
			goto not_valid;
		}
	} else {
		DBG_PRINT("No TCK needed.\n");
	}

	return(0);

reset_needed:
	return(1);

not_valid:
	return(-1);
}
#endif // #ifdef PARSE_ATR

/* Activate scard */
static int activate_scard(struct scard_private *priv)
{
	unsigned long old_jiffies, flags;
	
	unsigned char More_Delay = 0, ATR_Fail_CNT = 0;
	
	int ret = -1, XValue = 100;

	DBG_X_PRINT("[%s v%s] %s: scard activation starts ..[Tx_Mode=%d]\n", DRIVER_DATE,DRIVER_VERSION, priv->devname,Tx_Mode);

	spin_lock_irqsave(&priv->lock, flags);

	IRD_ATR_X_mode = 0;

RE_ACTIVE_CARD:

#ifdef SCARD_DEBUG_X1
	ISR_DBG_Count = 0;
#endif

#ifdef SCARD_DEBUG_X1
	if(ATR_Fail_CNT != 0)	DBG_X_PRINT(">>>>> Jump RE_ACTIVE_CARD...\n");
#endif

#ifdef CLASS_B // For 3V	
	if (priv->volt3 != 0) {
		DBG_PRINT("%s: Try 3V ..\n", priv->devname);
		scard_ctrlpin_write(priv->idx, priv->scard_5v_pin, 0); /* Use 3V */
		scard_ctrlpin_write(priv->idx, priv->scard_cmd_pin, 0); /* Drive CMDVCC low */
		udelay(RST_WAIT);

		/* Drive RSTIN low */
//		gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) & ~0x20);
//		gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) & ~0x620);
		gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) & ~0x6620);
		udelay(RST_WIDTH);
		gbus_write_uint32(pGBus, SCARD_INTEN_REG(priv->base_addr), 0x3f); /* Re-enable interrupt */
//		gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) | 0x20);
//		gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) | 0x620);
		gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) | 0x6620);
		gbus_write_uint32(pGBus, SCARD_STATE_REG(priv->base_addr), (unsigned long)PWR_IDLE);

		spin_unlock_irqrestore(&priv->lock, flags);

		for (old_jiffies = jiffies; time_after(old_jiffies + (HZ / 2), jiffies);) 
			/* Wait for activity from SCARD (max. 0.5sec)*/ ;

		gbus_write_uint32(pGBus, SCARD_INTEN_REG(priv->base_addr), 0); /* disable interrupt */
		spin_lock_irqsave(&priv->lock, flags);

		if (atomic_read(&(priv->answer2reset)) != 0) { 
#ifdef SCARD_DEBUG
			int i = 0;
			DBG_PRINT("Got response from 3V setting ..[priv->ans_idx = %d]\nAnswering: ",priv->ans_idx);
			for (i = 0; i < priv->ans_idx; i++)
				DBG_PRINT("0x%02x ", priv->ans_string[i]);
			DBG_PRINT("\n"); 
#endif
			if ((ret = parse_ans2reset(priv, 1, priv->ans_string, priv->ans_idx)) < 0) {
				DBG_PRINT("Ans2reset: string error\n");
				scard_reset(priv, 1); /* cold reset */
			} else if (ret > 0) {
				/* Perform warm reset */
				DBG_PRINT("Ans2reset: warm reset needed.\n");
				gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) & ~0x40);
				gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
							gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) | (1 << 2));
				scard_reset(priv, 0); /* warm reset */
				goto out;
			} else {
				/* The card is activated so we can turn on interrupts */
				gbus_write_uint32(pGBus, SCARD_INTEN_REG(priv->base_addr), 0x3f); /* Re-enable interrupt */
			}
		}
	}
#else	
	priv->volt3 = 0;	
#endif //#ifdef CLASS_B

	if ((priv->volt3 == 0) || (atomic_read(&(priv->answer2reset)) == 0)) {	/* 3V doesn't get response */
		int i;

		scard_ctrlpin_write(priv->idx, priv->scard_cmd_pin, 1); /* Drive CMDVCC high to deactive */
		for(i=0;i<50;i++)	udelay(1000);/// org -> udelay(1000);

		DBG_X_PRINT("%s: Try 5V ..\n", priv->devname);
		scard_ctrlpin_write(priv->idx, priv->scard_5v_pin, 1); /* Use 5V */
		for(i=0;i<50;i++)	udelay(1000);//add 2k90106
		scard_ctrlpin_write(priv->idx, priv->scard_cmd_pin, 0); /* Drive CMDVCC low */
	#if 1 // org
		udelay(RST_WAIT);	
	#else // test
		{
			unsigned int SafeCount;

			SafeCount = 0x3000;
        	while(SafeCount>0)
        	{
         	SafeCount--;
        	}
		}
	#endif
	
		/* Drive RSTIN low */
		#if 0 // ORG
//		gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) & ~0x20);
//		gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) & ~0x620);
		gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) & ~0x6620);
		udelay(RST_WIDTH);
		gbus_write_uint32(pGBus, SCARD_INTEN_REG(priv->base_addr), 0x3f); /* Re-enable interrupt */
//		gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) | 0x20);
//		gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) | 0x620);
		gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) | 0x6620);
		#else // by skkim 2k80705
		if(Tx_Mode == 2) // T=14
		{
			gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) & ~0x7F20);
			for(i=0;i<XValue;i++)	udelay(RST_WIDTH);
			gbus_write_uint32(pGBus, SCARD_INTEN_REG(priv->base_addr), 0x3f); // Re-enable interrupt
			gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) | 0x20);
		}
		else if(Tx_Mode == 3) // T=1 for Nagra..
		{
			gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) & ~0x7F20);
			for(i=0;i<XValue;i++)	udelay(RST_WIDTH);
			gbus_write_uint32(pGBus, SCARD_INTEN_REG(priv->base_addr), 0x3f); // Re-enable interrupt
			gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) | 0x20);
		}
		else
		{
			gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) & ~0x7F20);
			for(i=0;i<XValue;i++)	udelay(RST_WIDTH);
			gbus_write_uint32(pGBus, SCARD_INTEN_REG(priv->base_addr), 0x3f); // Re-enable interrupt
			gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) | 0x6620);
		}	
		#endif
		
		gbus_write_uint32(pGBus, SCARD_STATE_REG(priv->base_addr), (unsigned long)PWR_IDLE);
		
		spin_unlock_irqrestore(&priv->lock, flags);

		#if 0
		for (old_jiffies = jiffies; time_after(old_jiffies + (HZ / 2), jiffies);) 
			/* Wait for activity from SCARD (max. 0.5sec)*/ ;
		#else // by skkim 2k80703
	#if 0
		if(More_Delay == 1)
		{
			for (old_jiffies = jiffies; time_after(old_jiffies + (HZ), jiffies);) 
			/* Wait for activity from SCARD (max. 1.0sec)*/ ;
		}
		else if(More_Delay == 2)
		{
			for (old_jiffies = jiffies; time_after(old_jiffies + (HZ*2), jiffies);) 
			/* Wait for activity from SCARD (max. 2.0sec)*/ ;
		}
		else
		{
			for (old_jiffies = jiffies; time_after(old_jiffies + (HZ / 2), jiffies);) 
			/* Wait for activity from SCARD (max. 0.5sec)*/ ;
		}	
	#else // 2k90106
		if(More_Delay == 1)
		{
			for (old_jiffies = jiffies; time_after(old_jiffies + HZ+(HZ/2), jiffies);) 
			/* Wait for activity from SCARD (max. 1.5sec)*/ ;
		}
		else if(More_Delay == 2)
		{
			for (old_jiffies = jiffies; time_after(old_jiffies + (HZ*2)+(HZ/2), jiffies);) 
			/* Wait for activity from SCARD (max. 2.5sec)*/ ;
		}
		else if(More_Delay == 3)
		{
			for (old_jiffies = jiffies; time_after(old_jiffies + (HZ*3), jiffies);) 
			/* Wait for activity from SCARD (max. 3sec)*/ ;
		}
		else
		{
			for (old_jiffies = jiffies; time_after(old_jiffies + (HZ), jiffies);) 
			/* Wait for activity from SCARD (max. 1.0sec)*/ ;
		}	
	#endif
		#endif
		
		gbus_write_uint32(pGBus, SCARD_INTEN_REG(priv->base_addr), 0); /* disable interrupt */
		spin_lock_irqsave(&priv->lock, flags);
	
		if (atomic_read(&(priv->answer2reset)) != 0) { 
#ifdef SCARD_DEBUG_X1
			int i;
			DBG_X_PRINT("Got response from 5V setting ..[priv->ans_idx = %d]\nAnswering: ",priv->ans_idx);
			for (i = 0; i < priv->ans_idx; i++)
				DBG_X_PRINT("0x%02x ", priv->ans_string[i]);
			DBG_X_PRINT("\n"); 
#endif
				
			if((priv->ans_string[0] == 0x3B || priv->ans_string[0] == 0x3F)&&(priv->ans_idx < 4))
			{
				if(ATR_Fail_CNT < 2)
				{				
					/* Reset the RX/TX FIFOs */
					gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
								gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) | ((1 << 24) | (1 << 25)));
					gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
								gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) & ~((1 << 24) | (1 << 25)));
				}

				if((priv->ans_idx>1)&&(priv->ans_string[0]==0x3B)&&(priv->ans_string[1]==0x9F)&&((ATRXBuf[0]&0xF0)==0xC0))
				{
					XValue = 1000;

					if(ATR_Fail_CNT != 0)
						IRD_ATR_X_mode = 1;
				}
				else
				{
            	IRD_ATR_X_mode = 0;
				}

				if(ATR_Fail_CNT == 0)
				{
					ATR_Fail_CNT++;
					More_Delay = 1;
					priv->ans_idx = 0;

					goto RE_ACTIVE_CARD;
				}
				else if(ATR_Fail_CNT == 1)
				{
					ATR_Fail_CNT++;
					More_Delay = 2;
					priv->ans_idx = 0;

					goto RE_ACTIVE_CARD;
				}
				/*else if(ATR_Fail_CNT == 2)
				{
					ATR_Fail_CNT++;
					More_Delay = 3;
					priv->ans_idx = 0;

					goto RE_ACTIVE_CARD;
				}*/
			}
			else if((priv->ans_string[0] == 0x3F && priv->ans_string[1] == 0xFF && priv->ans_string[2] == 0x95
				 && priv->ans_string[3] == 0x00 && priv->ans_string[4] == 0xFF)/*&&((priv->ans_string[1]&0x0F)>priv->ans_idx)*/)
			{
				if(ATR_Fail_CNT == 0)
				{
					unsigned int nclkx;

					ATR_Fail_CNT = 2;
					Tx_Mode = 3;
					More_Delay = 1;
			 		priv->ans_idx = 0;

					nclkx = tangox_get_sysclock() / (3600000 * 2);
					DBG_X_PRINT(">> nclkx = %d | tangox_get_sysclock() = %ld\n", nclkx, tangox_get_sysclock());
					
					gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) & 0xffffffe0);
					gbus_write_uint32(pGBus, SCARD_CLK_HIGH_VAL(priv->base_addr), nclkx); 
					gbus_write_uint32(pGBus, SCARD_CLK_LOW_VAL(priv->base_addr), nclkx); 
		
					/* Reset the RX/TX FIFOs */
					gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
								gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) | ((1 << 24) | (1 << 25)));
					gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
								gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) & ~((1 << 24) | (1 << 25)));

					goto RE_ACTIVE_CARD;
				}	
			}
			
		#ifdef PARSE_ATR
			DBG_PRINT("-->> ATR Parsing Enable\n"); 
			
			if ((ret = parse_ans2reset(priv, 0, priv->ans_string, priv->ans_idx)) < 0) {
				DBG_PRINT("Ans2reset: string error\n");
				scard_reset(priv, 1); /* cold reset */
			} else if (ret > 0) {
				/* Perform warm reset */
				DBG_PRINT("Ans2reset: warm reset needed.\n");
				gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) & ~0x40);
				gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
							gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) | (1 << 2));
				
				scard_reset(priv, 0); /* warm reset */
				goto out;
			} else {
				/* The card is activated so we can turn on interrupts */
				gbus_write_uint32(pGBus, SCARD_INTEN_REG(priv->base_addr), 0x3f); /* Re-enable interrupt */
			}
		#else // by skkim 2k80703
			DBG_X_PRINT("-->> ATR Parsing Disable [priv->ans_idx = %d]\n",priv->ans_idx); 
			
			priv->rev_msb = priv->rev_pol = 0;

			if(ATRXBuf_CNT<5)	ATRXBuf[ATRXBuf_CNT] = priv->ans_string[0];

#ifdef SCARD_DEBUG_X1
			{
				int xx;
				DBG_X_PRINT("ATRXBuf = [\n"); 
				for(xx=0; xx<ATRXBuf_CNT+1; xx++)	DBG_X_PRINT(" %02X", ATRXBuf[xx]);
				DBG_X_PRINT("\n ]\n");
			}
#endif
			
			ATRXBuf_CNT++;
	
			switch(priv->ans_string[0]) {
				case 0x3b: /* Correct config detected */
				case 0x3f:
					DBG_X_PRINT("Correct ts detected 0x%x\n", priv->ans_string[0]);
					priv->rev_msb = priv->rev_pol = 0;

					if( (Tx_Mode ==2)&&(priv->ans_idx<20)&&(priv->ans_string[0]==0x3B)&&(priv->ans_string[1]==0x9F)&&((ATRXBuf[0]&0xF0)==0xC0) )
					{
						unsigned char IRD_Recover_ATR[20]={0x3B,0x9F,0x21,0x0E,0x49,0x52,0x44,0x45,0x54,0x4F,0x20,0x41,0x43,0x53,0x03,0x84,0x55,0xFF,0x80,0x6D};

						DBG_X_PRINT(" --->>> IRD_Recover_ATR\n");
					
						memcpy(priv->ans_string, IRD_Recover_ATR, 20);
						priv->ans_idx = 20;
					}
					break;
				case 0xc4:
				case 0xc0:
					DBG_X_PRINT("Wrong ts detected 0x%x (need to reverse the polarity)\n", priv->ans_string[0]);
					priv->rev_pol = 1;
					break;
				case 0xdc:
				case 0xfc:
					DBG_X_PRINT("Wrong ts detected 0x%x (need to reverse the MSB)\n", priv->ans_string[0]);
					priv->rev_msb = 1;
					break;
				case 0x23:
				case 0x03:
					DBG_X_PRINT("Wrong ts detected 0x%x (need to reverse the polarity and MSB)\n", priv->ans_string[0]);
					priv->rev_msb = priv->rev_pol = 1;
					break;

				case 0x73: // Add by AMZ 2k90127
					if((ATRXBuf[0]&0xF0)==0xC0) // T=14 Card...
					{
#ifdef SCARD_DEBUG_X1
						DBG_X_PRINT("Change to T=14 Mode..\n");
#endif
						priv->rev_msb = priv->rev_pol = 2;

						if(ATRXBuf[0] > 0xC5)	XValue = 200;  // for official Irdeto
						else							XValue = 1000; // for betacrypto or non official irdeto
						break;
					}
				
				case 0xEE:  // Add by AMZ 2k90127
				case 0xFF:  // Add by AMZ 2k90127
					if( ((ATRXBuf_CNT>2)&&(ATRXBuf[0]==0xFF)&&(ATRXBuf[1]==0xDC))      //T=0 SECA
							||	((ATRXBuf_CNT>2)&&(ATRXBuf[0]==0x77)&&(ATRXBuf[1]==0xDC)) //T=0 Dreamcrypt
						)
					{
#ifdef SCARD_DEBUG_X1
						DBG_X_PRINT("Change to T=0 Mode for rev_msb and rev_pol..\n");
#endif
						priv->rev_msb = priv->rev_pol = 1;
						XValue = 100;
						break;
					}
	
				default:
					if(Reset_mode_count++ > 8)
					{
						DBG_X_PRINT("Wrong ts detected 0x%x Linit Over[Reset_mode_count=%d]\n", priv->ans_string[0],Reset_mode_count);
						Reset_mode_count = 0;
						atomic_set(&(priv->normal_mode), 1);
						goto out_end;	
					}
					DBG_X_PRINT("Wrong ts detected 0x%x [Reset_mode_count=%d]\n", priv->ans_string[0],Reset_mode_count);
						
					if(Reset_mode_count < 4)
					{
						DBG_X_PRINT("priv->reset_cnt=%d\n", priv->reset_cnt);
						priv->reset_cnt++;
						if (priv->reset_cnt > 3)
							goto out;
						else {
							switch(priv->reset_cnt) {
								case 1: priv->rev_pol = (priv->rev_pol != 0) ? 0 : 1;
									break;
								case 2: priv->rev_msb = (priv->rev_msb != 0) ? 0 : 1;
									break;
								case 3: priv->rev_pol = (priv->rev_pol != 0) ? 0 : 1;
									priv->rev_msb = (priv->rev_msb != 0) ? 0 : 1;
									break;
							}
						}
						
						gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) & ~0x40);
						gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
							gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) | (1 << 2));
						scard_reset(priv, 0); /* warm reset */
						ret = 0;
						goto out;
					}
					else if(Reset_mode_count < 8)
					{
						priv->rev_msb = priv->rev_pol = 2;
					}	
					break;
			}
			
			if(priv->rev_msb == 1 || priv->rev_pol == 1)
			{
				gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) & ~0x40);
				gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
							gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) | (1 << 2));

				if((ATR_Fail_CNT == 2)&&(Tx_Mode == 2)) // 2k90109
				{
					ATR_Fail_CNT = Tx_Mode = 0;
				}

				scard_reset(priv, 0); /* warm reset */
				goto RE_ACTIVE_CARD;  // 2k90127
			}
			else if(priv->rev_msb == 2 || priv->rev_pol == 2)
			{
				/*priv->rev_msb = priv->rev_pol = 1;*/
				priv->rev_msb = priv->rev_pol = 0;

				gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) & ~0x40);
				gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
							gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) | (1 << 2));
				
				Tx_Mode = 2;
				
				scard_reset(priv, 3); /* warm reset with mode 3 */
				goto RE_ACTIVE_CARD;  // 2k90127
			}
			else
			{
				gbus_write_uint32(pGBus, SCARD_INTEN_REG(priv->base_addr), 0x3f); /* Re-enable interrupt */
			}
			goto out;		
		#endif	
		}
	}

out:
	if (atomic_read(&(priv->answer2reset)) != 0) {
//		gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) & ~0x600);
//		gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) | 0x600);
		gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) & ~0x6600);
		gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) | 0x6600);
		DBG_X_PRINT("%s: got response from the scard --> switch to normal\n", priv->devname);
		atomic_set(&(priv->normal_mode), 1);
		Reset_mode_count = Reset_Fail_count = 0;;
		ret = 1;
	} else if (ret < 0) {
		Reset_Fail_count++;
		DBG_X_PRINT("%s: no response from the scard --> reject[Reset_Fail_count = %d]\n", priv->devname,Reset_Fail_count);
		
		if(Reset_Fail_count > 5)
		{
			ERR_PRINT("%s: no response from the scard Limit Over..\n", priv->devname);
			Reset_mode_count = Reset_Fail_count = 0;
			gbus_write_uint32(pGBus, SCARD_INTEN_REG(priv->base_addr), 0); /* disable interrupt */
			atomic_set(&(priv->normal_mode), 1);
		}
		else
		{
			DBG_X_PRINT("%s: ATR_Fail_CNT = %d\n", priv->devname, ATR_Fail_CNT);
			if(ATR_Fail_CNT == 0)
			{
				ATR_Fail_CNT = 2;
				Tx_Mode = 2;
				More_Delay = 2;
				priv->ans_idx = 0;

				XValue = 1000;

				goto RE_ACTIVE_CARD;
			}

			if((Reset_mode_count == 0)&&(Reset_Fail_count == 2)) atomic_set(&(priv->normal_mode), 1);	
		}		
		
		/* by skkim :: scard_reset(priv, 1); */ /* cold reset */
		ret = 0;
	}
out_end:	
	spin_unlock_irqrestore(&priv->lock, flags);

#ifdef SCARD_DEBUG_X1
	DBG_X_PRINT(">>  activate_scard End.. ISR_DBG_Count = %d\n",ISR_DBG_Count);
#endif

	return(ret);
}

/* Reading from driver's buffer, note that it can return read size
   less than specified */
static int scard_read(struct file *fptr, char *buffer, size_t count, loff_t *fp)
{
	struct scard_private *priv = (struct scard_private *)fptr->private_data;
	unsigned long flags;
	unsigned char buf[rd_buffer_size], *bufptr = &buf[0];
	int cnt, i;
	
	///printk("scard_read :: size  = %d\n",count);
	
	if (atomic_read(&(priv->card_existed)) == 0)
		return(-EIO);
	else if (atomic_read(&(priv->normal_mode)) == 0)
		return(-EIO);

	/* If block mode is on, check the emptiness of buffer */
	if (priv->b_mode != 0) {
#ifdef READ_TIMEOUT
		Read_act = 1;
		Read_act_count = 0;
#endif
		/* Sleep when buffer is empty */
		wait_event_interruptible(priv->scard_rd_wq, priv->rd_c_idx != priv->rd_p_idx);

#ifdef READ_TIMEOUT
		if(Read_act == 2)
		{
			ERR_PRINT("%s: WARNING: recover Read_act_count  %d [read_cnt = %ld]\n", priv->devname, Read_act_count, priv->read_cnt);
			priv->rd_p_idx = Read_act_count;
		}	
		Read_act = 0;
		Read_act_count = 0;
#endif
		if (atomic_read(&(priv->card_existed)) == 0) 
			return(-EIO);
		else if (atomic_read(&(priv->normal_mode)) == 0) 
			return(-EIO);
	}

#if 1 // 2k90131
	if(count != 0)
	{
		/*int i;*/
		unsigned int time_count = 0;
		unsigned char read_value = 0;

		while(time_count < (RW_TIMEOUT+RW_MORE_DELAY))
		{
			if(priv->read_cnt >= count)
			{
				break;
			}

			time_count++;

			if(RW_Abort == 1)
			{
				return(0);
			}

			if (atomic_read(&(priv->card_existed)) == 0)
				return(-EIO);
			else if (atomic_read(&(priv->normal_mode)) == 0)
				return(-EIO);

			for(i=0;i<5;i++)	udelay(1000);
		}

		if(time_count == (RW_TIMEOUT+RW_MORE_DELAY))
		{
#ifdef SCARD_DEBUG_X1
			int i;

			DBG_X_PRINT(">> Read Time out[%d] :: read_cnt = %ld[0x%02lx]\n",time_count,priv->read_cnt,priv->read_cnt);

			DBG_X_PRINT("Err with read [ ");
			for(i=0; i<priv->read_cnt; i++)
				DBG_X_PRINT("%02X ",priv->rd_buffer[priv->rd_c_idx+i]);
			DBG_X_PRINT(" ]\n");
#endif
	
#if 1
			read_value = (unsigned char)priv->read_cnt;

			priv->read_cnt = 0;
			return((int)read_value);
#else
			spin_lock_irqsave(&priv->lock, flags);
   		
			/* Get the data out and adjust consumer index */
			for (cnt = 0; (priv->rd_c_idx != priv->rd_p_idx) && (cnt < count); cnt++) {
				*bufptr = priv->rd_buffer[priv->rd_c_idx];
				priv->rd_c_idx = (priv->rd_c_idx + 1) % rd_buffer_size;
				bufptr++;
			}
   		
			spin_unlock_irqrestore(&priv->lock, flags);
			
			/* Get the data to user */
			if (cnt && copy_to_user(buffer, (char *)&buf[0], cnt)) 
				return(-EFAULT);
   		
			return(cnt);
#endif
		}

		if((ISO_Non_T0_Act == 0)&&(count >= 1)&&(header_read==1)) // 2k90218 Check SW
		{
			unsigned int mcnt=1;

			header_read = 2;
			read_value = priv->rd_buffer[priv->rd_c_idx+count];
			DBG_X_PRINT(">> SW1 read_value = 0x%02X\n",read_value);

			if(((read_value&0xF0) == 0x90)||((read_value&0xF0) == 0x60))
			{
				int seek = 0, seek_status = 0;

				if(read_value == 0x60)
				{
					mcnt = 2+count;
		
					for(seek=1; seek<priv->read_cnt; seek++) //2k90707
					{
						if(priv->rd_buffer[priv->rd_c_idx+count+seek]==0x60)
						{
							mcnt += 1;
						}
					}
				}
				else 							
					mcnt = 1+count;

				time_count = 0;
				while(time_count < ((RW_TIMEOUT+RW_MORE_DELAY)/3))
				{
					if(priv->read_cnt >= mcnt)
					{
						break;
					}
					time_count++;
 
					if(RW_Abort == 1)
					{
						return(0);
					}

					if(read_value == 0x60)	
					{
						for(i=0;i<50;i++)	udelay(1000);

						for(seek=1;seek<16;seek++)
						{
							if((priv->rd_buffer[priv->rd_c_idx+count+seek]&0xF0)==0x90)
							{
								seek_status = 1;
								break;
							}
						#if 1 // for Error Status
							else if((priv->rd_buffer[priv->rd_c_idx+count+seek]>=0x61)&&(priv->rd_buffer[priv->rd_c_idx+count+seek]<=0x6F))
							{
								seek_status = 2;
								break;
							}
						#endif
							else if(priv->rd_buffer[priv->rd_c_idx+count+seek]==0x60) for(i=0;i<100;i++)	udelay(1000);
						}
					}

					if((read_value == 0x60)&&(seek_status))
					{
#ifdef SCARD_DEBUG_X1
				#if 1 // Only Debug - Must be remove or disable
						DBG_X_PRINT("T=0> NULL[ ");
						//for(i=0;i<seek;i++)	DBG_X_PRINT("%02X ", priv->rd_buffer[priv->rd_c_idx+count+i]);
						for(i=0;i<priv->read_cnt;i++)	DBG_X_PRINT("%02X ", priv->rd_buffer[priv->rd_c_idx+count+i]);
						DBG_X_PRINT("]\n");
				#endif
						DBG_X_PRINT("T=0 with NULL>> SW[%d] %02X:%02X || priv->read_cnt=%ld\n",seek, priv->rd_buffer[priv->rd_c_idx+count+seek],priv->rd_buffer[priv->rd_c_idx+count+seek+1],priv->read_cnt);
#endif
						for(i=0;i<50;i++)	udelay(1000);
						break;
					}

					if (atomic_read(&(priv->card_existed)) == 0)
						return(-EIO);
					else if (atomic_read(&(priv->normal_mode)) == 0)
						return(-EIO);

					for(i=0;i<5;i++)	udelay(1000);
				}

				if(time_count == ((RW_TIMEOUT+RW_MORE_DELAY)/3))
				{
#ifdef SCARD_DEBUG_X1
					DBG_X_PRINT(">> Read Time out[%d] at chk SW :: priv->read_cnt = %ld[0x%02lx]||count=%d\n",time_count,priv->read_cnt,priv->read_cnt,count);
					DBG_X_PRINT(">> SW1 read_value = 0x%02X\n",read_value);
					read_value = priv->rd_buffer[priv->rd_c_idx+count+1];
					DBG_X_PRINT(">> SW2 read_value = 0x%02X\n",read_value);
#endif	
					read_value = (unsigned char)priv->read_cnt;
      		
					priv->read_cnt = 0;
					return((int)read_value);
				}

				read_value = priv->rd_buffer[priv->rd_c_idx+count+1];
				///DBG_X_PRINT(">> SW2 read_value = 0x%02X\n",read_value);
			}
		}
		else if((ISO_Non_T0_Act == 0)&&(count == 1)&&(header_read==0)) // ACK
		{
			if((read_value != T0_ACK_Master)&&(T0_ACK_Master != 0xFF)&&(T0_Last_Write_Size == 5))
			{
				if(priv->read_cnt>0)	
				{	
					priv->read_cnt--;
				}
				header_read = 1;
				for(i=0;i<50;i++)	udelay(1000);
			}
			else
			{
				header_read = 1;
			}
		}

		///DBG_X_PRINT(">> Read count=%d : priv->read_cnt=%ld : header_read=%d\n",count,priv->read_cnt,header_read);

		if(((count == 8)||(count == 3))&&(header_read==0)&&(ISO_Non_T0_Act == 1)) // For T=1/T=14
		{
			header_read = 1;
			read_value = priv->rd_buffer[priv->rd_c_idx+(count-1)];
			read_value += 1;

			time_count = 0;

			///DBG_X_PRINT(">> PayLoad read_value=%d\n",read_value);

			while(time_count < ((RW_TIMEOUT+RW_MORE_DELAY)+(read_value*5)))
			{
				if((priv->read_cnt-count) >= read_value)
				{
					break;
				}
				time_count++;

				if(RW_Abort == 1)
				{
					return(0);
				}

				if (atomic_read(&(priv->card_existed)) == 0)
					return(-EIO);
				else if (atomic_read(&(priv->normal_mode)) == 0)
					return(-EIO);

				for(i=0;i<5;i++)	udelay(1000);
			}

			if(time_count == ((RW_TIMEOUT+RW_MORE_DELAY)+(read_value*5)))
			{
#ifdef SCARD_DEBUG_X1		
				int i;

				DBG_X_PRINT(">> 2nd-Read Time out[%d] :: read_value = %d[0x%02x]\n",time_count,read_value,read_value);


				DBG_X_PRINT("Err with read [ ");
				for(i=0; i<priv->read_cnt; i++)
					DBG_X_PRINT("%02X ",priv->rd_buffer[priv->rd_c_idx+i]);
				DBG_X_PRINT(" ]\n");
#endif

#if 1
				priv->read_cnt = 0;
				return(-EFAULT);
#else
				spin_lock_irqsave(&priv->lock, flags);
   			
				/* Get the data out and adjust consumer index */
				for (cnt = 0; (priv->rd_c_idx != priv->rd_p_idx) && (cnt < count); cnt++) {
					*bufptr = priv->rd_buffer[priv->rd_c_idx];
					priv->rd_c_idx = (priv->rd_c_idx + 1) % rd_buffer_size;
					bufptr++;
				}
   			
				spin_unlock_irqrestore(&priv->lock, flags);
				
				/* Get the data to user */
				if (cnt && copy_to_user(buffer, (char *)&buf[0], cnt)) 
					return(-EFAULT);
   			
				return(cnt);
#endif

			}
		}

		if (atomic_read(&(priv->card_existed)) == 0)
			return(-EIO);
		else if (atomic_read(&(priv->normal_mode)) == 0)
			return(-EIO);
	}
#endif

	spin_lock_irqsave(&priv->lock, flags);

	/* Get the data out and adjust consumer index */
	for (cnt = 0; (priv->rd_c_idx != priv->rd_p_idx) && (cnt < count); cnt++) {
		*bufptr = priv->rd_buffer[priv->rd_c_idx];
		priv->rd_c_idx = (priv->rd_c_idx + 1) % rd_buffer_size;
		bufptr++;
	}

	spin_unlock_irqrestore(&priv->lock, flags);
	
	/* Get the data to user */
	if (cnt && copy_to_user(buffer, (char *)&buf[0], cnt)) 
		return(-EFAULT);

	return(cnt);
}

/* Writing to smart card device */
static int scard_write(struct file *fptr, const char *bufptr, size_t size, loff_t *fp)
{
	struct scard_private *priv = (struct scard_private *)fptr->private_data;
	unsigned long flags, cj;
	unsigned char buffer[wr_buffer_size];
	unsigned wr_pidx, wr_cidx;
	int count = 0;
	int i;

	// 2k90130
	memset(priv->rd_buffer, 0, rd_buffer_size); // 2k90707

	priv->read_cnt = 0;
	priv->rd_p_idx = priv->rd_c_idx = 0;

	if(ISO_Non_T0_Act != 0)
		header_read = 0;
	else if((ISO_Non_T0_Act == 0)&&(size==5))
		header_read = 0;

	// 2k90130

	write_ready = 1;

	///printk("scard_write :: size  = %d\n",size);
	
	if (atomic_read(&(priv->card_existed)) == 0) {
		printk("Card not there\n");
		return(-EIO);
	} else if (atomic_read(&(priv->normal_mode)) == 0) {
		printk("Card not ready\n");
		return(-EIO);
	}
	
	if (size > wr_buffer_size) {
		ERR_PRINT("%s: exceed max. write size (%d)\n", 
			priv->devname, wr_buffer_size);
		return(-EIO);
	} else if (copy_from_user(&buffer[0], bufptr, size))
		return(-EFAULT);

	spin_lock_irqsave(&priv->lock, flags);

	cj = jiffies;
	while ((priv->last_rd_jiffies != 0) && time_after_eq(priv->last_rd_jiffies + TIMEOUT, cj)) {
		spin_unlock_irqrestore(&priv->lock, flags);
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout((priv->last_rd_jiffies + TIMEOUT) - cj);
		spin_lock_irqsave(&priv->lock, flags);
		cj = jiffies;
	}

	wr_pidx = priv->wr_p_idx;	/* Save the old index before proceeding */
	wr_cidx = priv->wr_c_idx;

	for (count = 0; count < size; count++) {
		///DBG_PRINT("%s: write data 0x%x\n", priv->devname, buffer[count]);
		/* Save it to buffer */
		if (((priv->wr_p_idx + 1) % wr_buffer_size) == wr_cidx) {
			ERR_PRINT("%s: write buffer full\n", priv->devname);
			goto out;
		}

		priv->wr_buffer[priv->wr_p_idx] = buffer[count];
		priv->wr_p_idx = (priv->wr_p_idx + 1) % wr_buffer_size;

	}

out:
	spin_unlock_irqrestore(&priv->lock, flags);

	/* Buffer was empty, trigger the writer */
	if (wr_cidx == wr_pidx) 
	{
		scard_write_device(priv);
	}
	else if(write_ready == 1) // 2k90210
	{
		unsigned int time_count = 0;

		while(time_count < (RW_TIMEOUT+RW_MORE_DELAY))
		{
			if(write_ready == 0)
			{
				break;
			}
			time_count++;

			if(RW_Abort == 1)
			{
				return(0);
			}

			if (atomic_read(&(priv->card_existed)) == 0)
				return(-EIO);
			else if (atomic_read(&(priv->normal_mode)) == 0)
				return(-EIO);

			for(i=0;i<5;i++)	udelay(1000);
		}

		if(time_count == (RW_TIMEOUT+RW_MORE_DELAY))
		{
#ifdef SCARD_DEBUG_X1		
			DBG_X_PRINT(">> Write Time out[%d]\n",time_count);
#endif
			return(-EFAULT);
		}
	}

	T0_Last_Write_Size = size;

	if((ISO_Non_T0_Act == 0)&&(size==5))	
	{
		T0_ACK_Master = buffer[1];
		for(i=0; i<5; i++)	udelay(1000); // Guard Delay
	}
	else												
	{
		T0_ACK_Master = 0xFF;
		for(i=0; i<50; i++)	udelay(1000); // Guard Delay
	}

	return(count);
}


static void scard_set_param(struct scard_private *priv, unsigned char *Mode_Buf)
{
	unsigned char Mode, Over_Clock;
	unsigned long mask;
	unsigned int nclk;
	int i;

	gbus_write_uint32(pGBus, SCARD_INTEN_REG(priv->base_addr), 0); /* disable interrupt */
	
	Mode = Mode_Buf[1];
	Over_Clock = Mode_Buf[2];

	if(Over_Clock == 0)
	{
		if (Mode == 0)
			frequency = 3571200; //3571200 / 3600000
		else if (Mode == 3)
			frequency = 6000000;	
		else if (Mode == 4)
			frequency = 3686400;//3686400;///3686000;	
		else
			frequency = 3600000;
	}
	else
	{
		if (Mode == 0)
			frequency = 5000000;
		else if (Mode == 3)
			frequency = 6000000;	
		else if (Mode == 4)
			frequency = 3686400;
		else
			frequency = 5000000;
	}
		    
	/* Initialize everything to be in known state */
	nclk = tangox_get_sysclock() / (frequency * 2);
	DBG_X_PRINT("scard_set_param :: clk = %d | tangox_get_sysclock() = %ld | frequency = %d | Mode = %d | OC = %d\n", nclk, tangox_get_sysclock(), frequency, Mode, Over_Clock);
	
	gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) & 0xffffffe0);
	gbus_write_uint32(pGBus, SCARD_CLK_HIGH_VAL(priv->base_addr), nclk); 
	gbus_write_uint32(pGBus, SCARD_CLK_LOW_VAL(priv->base_addr), nclk); 
	
	//////gbus_write_uint32(pGBus, SCARD_TIMEOUT_LOAD(priv->base_addr), 3000000);
	
	ISO_Non_T0_Act = 0; // 2k90201
	
	if (Mode == 2) // Viaccess
	{
		ISO_Non_T0_Act = 0; // 2k90201

		DBG_X_PRINT("Set param..>> Last_rev_msb = %d || Last_rev_pol = %d\n",Last_rev_msb, Last_rev_pol);
		
		gbus_write_uint32(pGBus, SCARD_PARAM_REG(priv->base_addr), (3 << 20) | (8 << 16) | (2 << 12) | (0x9c << 0));
		gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), ((2 - 1) << 16) | ((372 - 1) << 0));		
		
		mask = 0;
		mask = (1 << 5) | (1 << 6) | (1 << 8) | (3 << 9) | (3 << 13) | (1 << 16) | 
			(2 << 17) | (1 << 20) | (1 << 24) | (1 << 25) | (1 << 27) | (1 << 7) | (1 << 12);

		if (Last_rev_msb != 0)
			mask ^= (1 << 7);
		if (Last_rev_pol != 0)
			mask ^= (1 << 12);
		
		gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), mask);
		
		gbus_write_uint32(pGBus, SCARD_TIMEOUT_LOAD(priv->base_addr), 3000000);
	}
	else if (Mode == 3) // Irdeto
	{
		ISO_Non_T0_Act = 1; // 2k90201

		gbus_write_uint32(pGBus, SCARD_PARAM_REG(priv->base_addr), (3 << 20) | (8 << 16) | (2 << 12) | (0x9c << 0));
		gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), ((4/*1*/ - 1) << 16) | ((620 - 1) << 0));
		
		mask = 0;
		mask = (1 << 5) | (0 << 6) | (0 << 7) | (1 << 8) | (3 << 9) | (0 << 12) | (3 << 13) | (0 << 16) | 
			(1 << 17) | (1 << 20) | (1 << 24) | (1 << 25) | (0 << 27);
		
		gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), mask);
		
		gbus_write_uint32(pGBus, SCARD_TIMEOUT_LOAD(priv->base_addr), 3000000);
	}
	else if (Mode == 4) // Nagra Fi=9 / Di=5 : 512/16 = 32
	{	// Th = (3 * 256) = 768, Tg = 2048, Tb = 512, Tc = 39936
		ISO_Non_T0_Act = 1; // 2k90201

		gbus_write_uint32(pGBus, SCARD_PARAM_REG(priv->base_addr), (3 << 20) | (8 << 16) | (2 << 12) | (0x9c << 0));
		gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), ((2 - 1) << 16) | ((32 - 1) << 0));		

		mask = 0;
		mask = (1 << 5) | (0 << 6) | (1 << 7) | (1 << 8) | (3 << 9) | (1 << 12) | (3 << 13) | (0 << 16) | 
			(1 << 17) | (1 << 20) | (1 << 24) | (1 << 25) | (1 << 27) | (0 << 29) | (0 << 30);
		
		gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), mask);
		
		gbus_write_uint32(pGBus, SCARD_TIMEOUT_LOAD(priv->base_addr), 3000000);
	}
	else if (Mode == 5) // Dream Fi=1 / Di=3 : 372/4 = 93
	{
		ISO_Non_T0_Act = 1; // 2k90201

		gbus_write_uint32(pGBus, SCARD_PARAM_REG(priv->base_addr), (3 << 20) | (8 << 16) | (2 << 12) | (0x9c << 0));
		gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), ((2 - 1) << 16) | ((93 - 1) << 0));		
		
		mask = 0;
		mask = (1 << 5) | (0 << 6) | (1 << 8) | (3 << 9) | (3 << 13) | (0 << 16) | 
			(2 << 17) | (1 << 20) | (1 << 24) | (1 << 25) | (1 << 27) | (0 << 7) | (0 << 12);
		
		gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), mask);
		
		gbus_write_uint32(pGBus, SCARD_TIMEOUT_LOAD(priv->base_addr), 3000000);
	}
	else if (Mode == 6) // CryptoWorks
	{
		ISO_Non_T0_Act = 0; // 2k90201

		DBG_X_PRINT("Set param..>> Last_rev_msb = %d || Last_rev_pol = %d\n",Last_rev_msb, Last_rev_pol);
		
		gbus_write_uint32(pGBus, SCARD_PARAM_REG(priv->base_addr), (3 << 20) | (8 << 16) | (2 << 12) | (0x9c << 0));
		gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), ((2 - 1) << 16) | ((372 - 1) << 0));		
		
		mask = 0;
		mask = (1 << 5) | (1 << 6) | (1 << 8) | (3 << 9) | (3 << 13) | (1 << 16) | 
			(2 << 17) | (1 << 20) | (1 << 24) | (1 << 25) | (1 << 27) | (1 << 7) | (1 << 12);

		if (Last_rev_msb != 0)
			mask ^= (1 << 7);
		if (Last_rev_pol != 0)
			mask ^= (1 << 12);
		
		gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), mask);
		
		gbus_write_uint32(pGBus, SCARD_TIMEOUT_LOAD(priv->base_addr), 3000000);
	}
	else if (Mode == 7) // Conax
	{
		ISO_Non_T0_Act = 0; // 2k90201

		DBG_X_PRINT("Set param..>> Last_rev_msb = %d || Last_rev_pol = %d\n",Last_rev_msb, Last_rev_pol);
		
		gbus_write_uint32(pGBus, SCARD_PARAM_REG(priv->base_addr), (3 << 20) | (8 << 16) | (2 << 12) | (0x9c << 0));
		gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), ((2 - 1) << 16) | ((372 - 1) << 0));		
		
		mask = 0;
		mask = (1 << 5) | (1 << 6) | (1 << 8) | (3 << 9) | (3 << 13) | (1 << 16) | 
			(2 << 17) | (1 << 20) | (1 << 24) | (1 << 25) | (1 << 27) | (1 << 7) | (1 << 12);

		if (Last_rev_msb != 0)
			mask ^= (1 << 7);
		if (Last_rev_pol != 0)
			mask ^= (1 << 12);
		
		gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), mask);
		
		gbus_write_uint32(pGBus, SCARD_TIMEOUT_LOAD(priv->base_addr), 3000000);
	}	
	else if (Mode == 8) // NDS-372
	{
		ISO_Non_T0_Act = 0;

		DBG_X_PRINT("Set param..>> Last_rev_msb = %d || Last_rev_pol = %d\n",Last_rev_msb, Last_rev_pol);
		
		gbus_write_uint32(pGBus, SCARD_PARAM_REG(priv->base_addr), (3 << 20) | (8 << 16) | (2 << 12) | (0x9c << 0));
		gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), ((2 - 1) << 16) | ((372 - 1) << 0));		
		
		mask = 0;
		mask = (1 << 5) | (1 << 6) | (1 << 8) | (3 << 9) | (3 << 13) | (1 << 16) | 
			(2 << 17) | (1 << 20) | (1 << 24) | (1 << 25) | (1 << 27) | (1 << 7) | (1 << 12);

		if (Last_rev_msb != 0)
			mask ^= (1 << 7);
		if (Last_rev_pol != 0)
			mask ^= (1 << 12);
		
		gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), mask);
		
		gbus_write_uint32(pGBus, SCARD_TIMEOUT_LOAD(priv->base_addr), 3000000);
	}
	else if (Mode == 9) // NDS-93
	{
		ISO_Non_T0_Act = 0;

		DBG_X_PRINT("Set param..>> Last_rev_msb = %d || Last_rev_pol = %d\n",Last_rev_msb, Last_rev_pol);
		
		gbus_write_uint32(pGBus, SCARD_PARAM_REG(priv->base_addr), (3 << 20) | (8 << 16) | (2 << 12) | (0x9c << 0));
		gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), ((4/*3*/ - 1) << 16) | ((93 - 1) << 0));		
		
		mask = 0;
		mask = (1 << 5) | (1 << 6) | (1 << 8) | (3 << 9) | (3 << 13) | (1 << 16) | 
			(2 << 17) | (1 << 20) | (1 << 24) | (1 << 25) | (1 << 27) | (1 << 7) | (1 << 12);

		if (Last_rev_msb != 0)
			mask ^= (1 << 7);
		if (Last_rev_pol != 0)
			mask ^= (1 << 12);

		gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), mask);
		
		gbus_write_uint32(pGBus, SCARD_TIMEOUT_LOAD(priv->base_addr), 3000000);
	}
	else if (Mode == 10) // NDS-46.5
	{
		ISO_Non_T0_Act = 0;

		DBG_X_PRINT("Set param..>> Last_rev_msb = %d || Last_rev_pol = %d\n",Last_rev_msb, Last_rev_pol);
		
		gbus_write_uint32(pGBus, SCARD_PARAM_REG(priv->base_addr), (3 << 20) | (8 << 16) | (2 << 12) | (0x9c << 0));
		gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), ((2 - 1) << 16) | ((46 - 1) << 0));		
		
		mask = 0;
		mask = (1 << 5) | (1 << 6) | (1 << 8) | (3 << 9) | (3 << 13) | (1 << 16) | 
			(2 << 17) | (1 << 20) | (1 << 24) | (1 << 25) | (1 << 27) | (1 << 7) | (1 << 12);

		if (Last_rev_msb != 0)
			mask ^= (1 << 7);
		if (Last_rev_pol != 0)
			mask ^= (1 << 12);
		
		gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), mask);
		
		gbus_write_uint32(pGBus, SCARD_TIMEOUT_LOAD(priv->base_addr), 3000000);
	}
	else
	{
		ISO_Non_T0_Act = 0; // 2k90201

		DBG_X_PRINT("Set param..>> Last_rev_msb = %d || Last_rev_pol = %d\n",Last_rev_msb, Last_rev_pol);
		
		gbus_write_uint32(pGBus, SCARD_PARAM_REG(priv->base_addr), (3 << 20) | (8 << 16) | (2 << 12) | (0x9c << 0));
		gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), ((2 - 1) << 16) | ((372 - 1) << 0));		
		
		mask = 0;
		mask = (1 << 5) | (1 << 6) | (1 << 8) | (3 << 9) | (3 << 13) | (1 << 16) | 
			(2 << 17) | (1 << 20) | (1 << 24) | (1 << 25) | (1 << 27) | (1 << 7) | (1 << 12);

		if (Last_rev_msb != 0)
			mask ^= (1 << 7);
		if (Last_rev_pol != 0)
			mask ^= (1 << 12);
		
		gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), mask);
		
		gbus_write_uint32(pGBus, SCARD_TIMEOUT_LOAD(priv->base_addr), 3000000);
	}	
	
	for(i=0; i<50; i++)	udelay(1000);
	
	/* Un-reset FIFO */
	gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
				gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) & ~((1 << 24) | (1 << 25)));
	
	gbus_write_uint32(pGBus, SCARD_INTEN_REG(priv->base_addr), 0x3f); /* Re-enable interrupt */
}


/*
Mode_Buff[0] // Card_Nop		Not used
Mode_Buff[1] // Card_Mode		0..2:Common / 3:T=14 for Irdeto / 4:T=1 for Nagra / 5:T=1 for Dream
Mode_Buff[2] // Card_Type		0:T=0 / 1:T=1 / 2:T=14

Mode_Buff[3] // Card_Clock_H	Clock Value of High
Mode_Buff[4] // Card_Clock_L	Clock Value of Low
										SmartCard_Clock = frequency = ((Card_Clock_H<<8)|Card_Clock_L)*100 [Unit:Hz]

Mode_Buff[5] // Card_FI			fi_tab[Card_FI]
Mode_Buff[6] // Card_DI			di_tab[Card_DI]
	static const unsigned fi_tab[16] = { 372, 372, 558, 744, 1116, 1488, 1860, 0, 0, 512, 768, 1024, 1536, 2048, 0, 0 };
	static const unsigned di_tab[16] = { 0, 1, 2, 4, 8, 16, 32, 0, 12, 20, 0, 0, 0, 0, 0, 0 };

Mode_Buff[7] // Card_EGT			0 ~ 4
Mode_Buff[8] // Card_TX_Error	0:Disabel / 1:Enable 
Mode_Buff[9] // Card_RX_Error	0:Disabel / 1:Enable 
Temp_buf[10] // TX Wait	Default 2 :: Max to 7 
Temp_buf[11] // Extra RW Delay Value x 100; Value x 100 x 5mS
Temp_buf[12] // TX/RX Retry ID MAX 7
*/

static void scard_set_dynamic_param(struct scard_private *priv, unsigned char *Mode_Buff)
{
	unsigned long mask;
	unsigned int nclk;
	unsigned char Mode, Card_Type, Card_Clock_H, Card_Clock_L, Card_FI, Card_DI, Card_EGT, Card_TX_Error, Card_RX_Error, Card_TX_Wait, Card_RTxRETRY;
	unsigned int Card_ETU = 371;
	int i;

	gbus_write_uint32(pGBus, SCARD_INTEN_REG(priv->base_addr), 0); /* disable interrupt */

	Mode 		 	  = Mode_Buff[1];
	Card_Type 	  = Mode_Buff[2];
	Card_Clock_H  = Mode_Buff[3];
	Card_Clock_L  = Mode_Buff[4];
	Card_FI		  = Mode_Buff[5];
	Card_DI		  = Mode_Buff[6];
	Card_EGT		  = Mode_Buff[7];
	Card_TX_Error = Mode_Buff[8];
	Card_RX_Error = Mode_Buff[9];
	Card_TX_Wait  = Mode_Buff[10];

	RW_MORE_DELAY = Mode_Buff[11];
	RW_MORE_DELAY = (RW_MORE_DELAY * 100);

	Card_RTxRETRY = Mode_Buff[12];

	DBG_X_PRINT("==[ scard_set_dynamic_param ] ==\n");
	DBG_X_PRINT("> Mode = %d | Card_Type = %d | Card_Clock_H = %d | Card_Clock_L = %d\n", Mode, Card_Type, Card_Clock_H, Card_Clock_L);
	DBG_X_PRINT("> Card_FI = %d | Card_DI = %d  | Card_EGT = %d \n", Card_FI, Card_DI, Card_EGT);
	DBG_X_PRINT("> Card_TX_Error = %d | Card_RX_Error = %d | Card_TX_Wait = %d\n", Card_TX_Error,Card_RX_Error,Card_TX_Wait);
	DBG_X_PRINT("==RW_MORE_DELAY = %d | Card_RTxRETRY = %d\n", RW_MORE_DELAY, Card_RTxRETRY);
	
	frequency = ((Card_Clock_H<<8)|Card_Clock_L)*100;
			    
	/* Initialize everything to be in known state */
	nclk = tangox_get_sysclock() / (frequency * 2);
	DBG_X_PRINT("scard_set_dynamic_param :: clk = %d | tangox_get_sysclock() = %ld\n", nclk, tangox_get_sysclock());
	DBG_X_PRINT("scard_set_dynamic_param :: frequency = %d | Mode = %d\n", frequency, Mode);
	
	gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) & 0xffffffe0);
	gbus_write_uint32(pGBus, SCARD_CLK_HIGH_VAL(priv->base_addr), nclk); 
	gbus_write_uint32(pGBus, SCARD_CLK_LOW_VAL(priv->base_addr), nclk); 
	
	//////gbus_write_uint32(pGBus, SCARD_TIMEOUT_LOAD(priv->base_addr), 3000000);
	
	ISO_Non_T0_Act = 0; // 2k90201

	Card_ETU = ((fi_tab[Card_FI] / di_tab[Card_DI]) - 1);
	DBG_X_PRINT("scard_set_dynamic_param..>> Card_ETU = %d\n",Card_ETU);
	DBG_X_PRINT("scard_set_dynamic_param..>> Last_rev_msb = %d || Last_rev_pol = %d\n",Last_rev_msb, Last_rev_pol);

	if (Mode == 3) // Irdeto
	{
		ISO_Non_T0_Act = 1; // 2k90201

		gbus_write_uint32(pGBus, SCARD_PARAM_REG(priv->base_addr), (3 << 20) | (8 << 16) | (2 << 12) | (0x9c << 0));
		gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), ((Card_EGT) << 16) | ((620 - 1) << 0));
		
		mask = 0;
		mask = (1 << 5) | (Card_TX_Error << 6) | (0 << 7) | (1 << 8) | (Card_RTxRETRY << 9) | (0 << 12) | (Card_RTxRETRY << 13) | (Card_RX_Error << 16) | 
			(Card_TX_Wait << 17) | (1 << 20) | (1 << 24) | (1 << 25) | (0 << 27);
		
		gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), mask);
		
		gbus_write_uint32(pGBus, SCARD_TIMEOUT_LOAD(priv->base_addr), 3000000);
	}
	else if (Mode == 4) // Nagra Fi=9 / Di=5 : 512/16 = 32
	{	// Th = (3 * 256) = 768, Tg = 2048, Tb = 512, Tc = 39936
		ISO_Non_T0_Act = 1; // 2k90201

		gbus_write_uint32(pGBus, SCARD_PARAM_REG(priv->base_addr), (3 << 20) | (8 << 16) | (2 << 12) | (0x9c << 0));
		gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), ((Card_EGT) << 16) | ((32 - 1) << 0));		

		mask = 0;
		mask = (1 << 5) | (Card_TX_Error << 6) | (1 << 7) | (1 << 8) | (Card_RTxRETRY << 9) | (1 << 12) | (Card_RTxRETRY << 13) | (Card_RX_Error << 16) | 
			(Card_TX_Wait << 17) | (1 << 20) | (1 << 24) | (1 << 25) | (1 << 27) | (0 << 29) | (0 << 30);
		
		gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), mask);
		
		gbus_write_uint32(pGBus, SCARD_TIMEOUT_LOAD(priv->base_addr), 3000000);
	}
	else if (Mode == 5) // Dream Fi=1 / Di=3 : 372/4 = 93
	{
		ISO_Non_T0_Act = 1; // 2k90201

		gbus_write_uint32(pGBus, SCARD_PARAM_REG(priv->base_addr), (3 << 20) | (8 << 16) | (2 << 12) | (0x9c << 0));
		gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), ((Card_EGT) << 16) | ((93 - 1) << 0));		
		
		mask = 0;
		mask = (1 << 5) | (Card_TX_Error << 6) | (1 << 8) | (Card_RTxRETRY << 9) | (Card_RTxRETRY << 13) | (Card_RX_Error << 16) | 
			(Card_TX_Wait << 17) | (1 << 20) | (1 << 24) | (1 << 25) | (1 << 27) | (0 << 7) | (0 << 12);
		
		gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), mask);
		
		gbus_write_uint32(pGBus, SCARD_TIMEOUT_LOAD(priv->base_addr), 3000000);
	}
	else if ((Mode&0xF0) == 0xF0) // common for Another case of T=1 or T=14
	{
		ISO_Non_T0_Act = 1;

		gbus_write_uint32(pGBus, SCARD_PARAM_REG(priv->base_addr), (3 << 20) | (8 << 16) | (2 << 12) | (0x9c << 0));
		gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), ((Card_EGT) << 16) | ((Card_ETU) << 0));		
		
		mask = 0;
		mask = (1 << 5) | (Card_TX_Error << 6) | (1 << 8) | (Card_RTxRETRY << 9) | (Card_RTxRETRY << 13) | (Card_RX_Error << 16) | 
			(Card_TX_Wait << 17) | (1 << 20) | (1 << 24) | (1 << 25) | (1 << 27) | (1 << 7) | (1 << 12);

		if (Last_rev_msb != 0)
			mask ^= (1 << 7);
		if (Last_rev_pol != 0)
			mask ^= (1 << 12);
		
		gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), mask);
		
		gbus_write_uint32(pGBus, SCARD_TIMEOUT_LOAD(priv->base_addr), 3000000);
	}
	else  // common for T=0
	{
		ISO_Non_T0_Act = 0;

		gbus_write_uint32(pGBus, SCARD_PARAM_REG(priv->base_addr), (3 << 20) | (8 << 16) | (2 << 12) | (0x9c << 0));
		gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), ((Card_EGT) << 16) | ((Card_ETU) << 0));		
		
		mask = 0;
		mask = (1 << 5) | (Card_TX_Error << 6) | (1 << 8) | (Card_RTxRETRY << 9) | (Card_RTxRETRY << 13) | (Card_RX_Error << 16) | 
			(Card_TX_Wait << 17) | (1 << 20) | (1 << 24) | (1 << 25) | (1 << 27) | (1 << 7) | (1 << 12);

		if (Last_rev_msb != 0)
			mask ^= (1 << 7);
		if (Last_rev_pol != 0)
			mask ^= (1 << 12);
		
		gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), mask);
		
		gbus_write_uint32(pGBus, SCARD_TIMEOUT_LOAD(priv->base_addr), 3000000);
	}	
	
	for(i=0; i<50; i++)	udelay(1000);
	
	/* Un-reset FIFO */
	gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
				gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) & ~((1 << 24) | (1 << 25)));
	
	gbus_write_uint32(pGBus, SCARD_INTEN_REG(priv->base_addr), 0x3f); /* Re-enable interrupt */
}
	
/* ioctl function */
static int scard_ioctl(struct inode *inode, struct file *fptr, unsigned int cmd, unsigned long arg)
{
	unsigned char *ptr = (unsigned char *)arg;
	struct scard_private *priv = (struct scard_private *)fptr->private_data;
	unsigned char result = 0;
	
	
	/////DBG_X_PRINT("->> scard_ioctl cmd = %d | ptr[0] = 0x%02x | ptr[1] = 0x%02x\n",cmd, ptr[0], ptr[1]);

#if 0 //org	
	if ( cmd == SCARD_IOC_CHECKCARD ) {
#else // 2k80712
	if ( (cmd == SCARD_IOC_CHECKCARD)&&(ptr[0] == 0x00 || ptr == NULL) ) {
#endif
		if (atomic_read(&(priv->card_existed)) == 1) {
			result |= 0x01;
		}

		if (atomic_read(&(priv->normal_mode)) == 1) {
			result |= 0x02;
		}

		return result;
	}


	if (atomic_read(&(priv->card_existed)) == 0)
		return(-EIO);
	else if (atomic_read(&(priv->normal_mode)) == 0)
		return(-EIO);
			
#if 1 // 2k80707
	if ( (cmd == SCARD_IOC_CHECKCARD)&&(ptr[0] == 0x01) ) // Req. ATR
	{
		if(atomic_read(&(priv->normal_mode)) == 1)
		{
			memcpy(ptr, priv->ans_string, priv->ans_idx);
			
			return priv->ans_idx;
		}
		
		return 0;
	}
	else if ( (cmd == SCARD_IOC_CHECKCARD)&&(ptr[0] == 0x02) ) // Req. Set parameter
	{
		scard_set_param(priv, &ptr[0]);
		
		return ptr[1];
	}
	else if ( (cmd == SCARD_IOC_WARMRESET)&&(ptr[0] == 0x03) ) // Req. Card Reset
	{
		scard_reset(priv, ptr[1]); /* Mode reset */
		
		return ptr[1];
	}	
	else if ( (cmd == SCARD_IOC_CHECKCARD)&&(ptr[0] == 0x04) ) // Req. Set dynamic parameter
	{
		scard_set_dynamic_param(priv, &ptr[0]);
		
		return ptr[1];
	}
	else if ( (cmd == SCARD_IOC_CHECKCARD)&&(ptr[0] == 0xF1)&&(ptr[1] == 0x01) ) // Req. Act fully close a card
	{
		unsigned long flags;

		RW_Abort = 1;

		Remove_Act = 0;
		Remove_Act_wait_count = 0;

		spin_lock_irqsave(&priv->lock, flags);	
		gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
					gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) | ((1 << 24) | (1 << 25)));
		gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
					gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) & ~((1 << 24) | (1 << 25)));

		if(priv->ref_cnt > 0)
			 priv->ref_cnt--;
		priv->rd_p_idx = priv->rd_c_idx = 0;
		priv->wr_p_idx = priv->wr_c_idx = 0;
		spin_unlock_irqrestore(&priv->lock, flags);

		MSG_PRINT("###################################################################################\n");
		MSG_PRINT("%s: fully close a card by Req... \n",priv->devname);
		MSG_PRINT("###################################################################################\n");

		RW_Abort = 0;
		
		return 1;
	}
#endif

	switch(cmd) {
		case SCARD_IOC_WARMRESET:
			gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) & ~0x40);
			gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
					gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) | (1 << 2));
			udelay(1);
			scard_reset(priv, 0); /* warm reset */
			break;
		case SCARD_IOC_CLOCKSTOP:
			if (priv->clockstop == 0) 
				return(-EIO);	
			gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
					gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) | (1 << 3));
			udelay(1);
			atomic_set(&(priv->state), (int)CLOCK_STOP);
			break;
		case SCARD_IOC_CLOCKSTART:
			if (priv->clockstop == 0) 
				return(-EIO);	
			gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
					gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) & ~(1 << 3));
			udelay(1);
			atomic_set(&(priv->state), (int)PWR_IDLE);
			break;
		default:
			return(-EIO);
	}
	return(0);
}

/* Poll function */
static unsigned int scard_poll(struct file *fptr, struct poll_table_struct *ptable)
{
	struct scard_private *priv = (struct scard_private *)fptr->private_data;
	unsigned int mask = 0;
	unsigned long flags;

	if (atomic_read(&(priv->card_existed)) == 0)
		return(POLLERR);
	else if (atomic_read(&(priv->normal_mode)) == 0)
		return(POLLERR);

	spin_lock_irqsave(&priv->lock, flags);
	poll_wait(fptr, &(priv->scard_rd_wq), ptable);
	if (priv->rd_c_idx != priv->rd_p_idx)
		mask |= (POLLIN | POLLRDNORM);
	spin_unlock_irqrestore(&priv->lock, flags);

	return(mask);
}

/* Open the device */
static int scard_open(struct inode *inode_ptr, struct file *fptr)
{
	int minor = MINOR(inode_ptr->i_rdev);

	if (minor >= NUM_SCARD)
		return(EIO);
	else if (atomic_read(&(scard_priv[minor].card_existed)) == 0)
		return(-EIO);
	else if (atomic_read(&(scard_priv[minor].normal_mode)) == 0)
		return(-EIO);
	
	/* This device is exclusive, that is, only one process can use it */
	if (scard_priv[minor].ref_cnt != 0) {
		
		#if 1
		ERR_PRINT("%s: exclusive access only < scard_priv[minor:%d].ref_cnt=%d >\n", scard_priv[minor].devname, minor, scard_priv[minor].ref_cnt);
		#endif
		
		return(-EIO);
	} 
	
	/* Set the block mode and increase reference count */
	scard_priv[minor].ref_cnt++;
	scard_priv[minor].b_mode = ((fptr->f_flags & O_NONBLOCK) ? 0 : 1);
	printk("scard_open :: scard_priv[minor=%d].b_mode  = %d\n",minor,scard_priv[minor].b_mode);
	
	scard_priv[minor].last_rd_jiffies = jiffies;
	scard_priv[minor].last_wr_jiffies = jiffies;

	/* Flush the buffer */
	scard_priv[minor].rd_p_idx = scard_priv[minor].rd_c_idx = 0;
	scard_priv[minor].wr_p_idx = scard_priv[minor].wr_c_idx = 0;

	fptr->f_op = &scard_fops;
	fptr->private_data = (void *)&scard_priv[minor];

	atomic_set(&scard_priv[minor].state, (int)PWR_IDLE);

	return(0);
}

/* Close the device */
static int scard_release(struct inode *inode_ptr, struct file *fptr) 
{
	struct scard_private *priv = (struct scard_private *)fptr->private_data;
	unsigned long flags;

#if 0 // 2k80711 :: Disable..
	if (atomic_read(&(priv->card_existed)) == 0)
		goto error_out;
	else if (atomic_read(&(priv->normal_mode)) == 0)
		goto error_out;
#endif

	spin_lock_irqsave(&priv->lock, flags);

	/* Reset the RX/TX FIFOs */
	gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
				gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) | ((1 << 24) | (1 << 25)));
	gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
				gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) & ~((1 << 24) | (1 << 25)));

#ifdef SCARD_DEBUG_X1
	DBG_X_PRINT("[scard_release] Now Currnet priv->ref_cnt = %d\n",priv->ref_cnt);
#endif

	/* Adjust reference count */
	if(priv->ref_cnt > 0) // ADD::2k80711 / Mod:2k90213
		 priv->ref_cnt--;

#ifdef SCARD_DEBUG_X1
	DBG_X_PRINT("[scard_release] After priv->ref_cnt = %d\n",priv->ref_cnt);
#endif
	
	priv->rd_p_idx = priv->rd_c_idx = 0;
	priv->wr_p_idx = priv->wr_c_idx = 0;

#if 1 // 2k90214
	{
		int minor = MINOR(inode_ptr->i_rdev); /// 2k90214
/* 2k90215
		if (minor < NUM_SCARD)	
		{
			if(scard_priv[minor].ref_cnt > 0)	scard_priv[minor].ref_cnt--;
		}
*/
#ifdef SCARD_DEBUG_X1
		DBG_X_PRINT("[scard_release] scard_priv[minor:%d].ref_cnt = %d\n",minor,scard_priv[minor].ref_cnt);
#endif
	}
#endif

	Remove_Act = 0;

	spin_unlock_irqrestore(&priv->lock, flags);

	return(0);

#if 0 // 2k80711 :: Disable..
error_out:
	/* Adjust reference count */
	priv->ref_cnt--;
	return(-EIO);
#endif	
}

/* Reset scard to initial state */
static void scard_reset(struct scard_private *priv, int cold)
{
	unsigned long mask;
	unsigned int nclk;

	gbus_write_uint32(pGBus, SCARD_INTEN_REG(priv->base_addr), 0); /* disable interrupt */

	#if 0
	if (cold != 0){ /* Cold reset */
	#else
	if (cold == 1){ /* Cold reset */
	#endif
		priv->volt3 = use_3v;
		priv->rev_msb = reverse_msb;
		priv->rev_pol = reverse_pol;
	}
	priv->ans_idx = 0;
	priv->clockstop = 0;
	priv->rd_p_idx = priv->rd_c_idx = 0;
	priv->wr_p_idx = priv->wr_c_idx = 0;
	
	if (cold == 5) // SECA, DREAMCRYPT
		priv->rev_pol = priv->rev_msb = 1;

	Last_rev_msb = priv->rev_msb;
	Last_rev_pol = priv->rev_pol;
	
	/* Set CMDVCC high and Set RSTIN high */
	scard_ctrlpin_write(priv->idx, priv->scard_cmd_pin, 1); 

	if (priv->volt3 != 0)
		scard_ctrlpin_write(priv->idx, priv->scard_5v_pin, 0); 
	else
		scard_ctrlpin_write(priv->idx, priv->scard_5v_pin, 1); 

	gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) | 0x60);
	
	// by skkim 2k80703
	if (cold == 3)
		frequency = 6000000;
	else if (cold == 4)
		frequency = 3600000;
	else
		frequency = 3600000; //2000000; ///3571200; //3571200 / 3600000
			    
	/* Initialize everything to be in known state */
	nclk = tangox_get_sysclock() / (frequency * 2);
	DBG_X_PRINT("clk = %d | tangox_get_sysclock() = %ld | frequency = %d | cold = %d\n", nclk, tangox_get_sysclock(), frequency, cold);
	
	gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) & 0xffffffe0);
	gbus_write_uint32(pGBus, SCARD_CLK_HIGH_VAL(priv->base_addr), nclk); 
	gbus_write_uint32(pGBus, SCARD_CLK_LOW_VAL(priv->base_addr), nclk); 

	#if 0
	/* Th = (3 * 256) = 768, Tg = 2048, Tb = 512, Tc = 39936 */
	gbus_write_uint32(pGBus, SCARD_PARAM_REG(priv->base_addr), (3 << 20) | (8 << 16) | (2 << 12) | (0x9c << 0));
	/* EGT = 4, ETU = 372 */
	gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), ((4 - 1) << 16) | ((372 - 1) << 0));


	/* set up initial value (includes resetting FIFO */
	mask = (1 << 5) | (1 << 6) | (1 << 8) | (3 << 9) | (3 << 13) | (1 << 16) | 
		(2 << 17) | (1 << 20) | (1 << 24) | (1 << 25) | (1 << 27) | (1 << 7) | (1 << 12);
	if (priv->rev_msb != 0)
		mask ^= (1 << 7);
	if (priv->rev_pol != 0)
		mask ^= (1 << 12);
	gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), mask);
	#else // 2k80703 by skkim
	if (cold == 3) // Irdeto
	{
		gbus_write_uint32(pGBus, SCARD_PARAM_REG(priv->base_addr), (3 << 20) | (8 << 16) | (2 << 12) | (0x9c << 0));
		gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), ((4/*1*/ - 1) << 16) | ((620 - 1) << 0));
		
		mask = 0;
		mask = (1 << 5) | (0 << 6) | (0 << 7) | (1 << 8) | (3 << 9) | (0 << 12) | (3 << 13) | (0 << 16) | 
			(1 << 17) | (1 << 20) | (1 << 24) | (1 << 25) | (0 << 27);
		
			/*
			mask ^= (1 << 6);
			mask ^= (1 << 7);
			mask ^= (3 << 9);
			mask ^= (1 << 12);
			mask ^= (3 << 13);
			mask ^= (1 << 16);
			mask ^= (1 << 27);
			
			mask |= (1 << 29);
			*/
			
		gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), mask);
		
		gbus_write_uint32(pGBus, SCARD_TIMEOUT_LOAD(priv->base_addr), 3000);
		
	}
	else
	{
		DBG_X_PRINT("Reset..>> priv->rev_msb = %d || priv->rev_pol = %d\n",priv->rev_msb, priv->rev_pol);
		
		gbus_write_uint32(pGBus, SCARD_PARAM_REG(priv->base_addr), (3 << 20) | (8 << 16) | (2 << 12) | (0x9c << 0));
		gbus_write_uint32(pGBus, SCARD_EGT_ETU_REG(priv->base_addr), ((4 - 1) << 16) | ((372 - 1) << 0));		

		mask = (1 << 5) | (1 << 6) | (1 << 8) | (3 << 9) | (3 << 13) | (1 << 16) | 
			(2 << 17) | (1 << 20) | (1 << 24) | (1 << 25) | (1 << 27) | (1 << 7) | (1 << 12);
		if (priv->rev_msb != 0)
			mask ^= (1 << 7);
		if (priv->rev_pol != 0)
			mask ^= (1 << 12);
		gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), mask);
		
		gbus_write_uint32(pGBus, SCARD_TIMEOUT_LOAD(priv->base_addr), 3000);
	}	
	#endif
	
	udelay(1000);//mod 2k90109 udelay(1);
	udelay(1000);//add 2k90109

	/* Un-reset FIFO */
	gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
				gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) & ~((1 << 24) | (1 << 25)));
	gbus_write_uint32(pGBus, SCARD_STATE_REG(priv->base_addr), 0);
	
	#if 0
	atomic_set(&(priv->card_existed), 0);
	atomic_set(&(priv->state), (int)UNKNOWN);
	#else // 2k80703
	if(cold == 1) 
	{
		atomic_set(&(priv->card_existed), 0);
		atomic_set(&(priv->state), (int)UNKNOWN);
	}
	#endif
	
	atomic_set(&(priv->normal_mode), 0);
	atomic_set(&(priv->answer2reset), 0);

//	gbus_write_uint32(pGBus, SCARD_INTEN_REG(priv->base_addr), 0x3f); /* Re-enable interrupt */
}

/* Monitor the status of smart card */
static int scard_monitor(void *priv_data)
{
	struct scard_private *priv = (struct scard_private *)priv_data;
	enum SCARD_STATE cur_state, set_state;
	int p_card_state;
	int c_card_state;
	int delay;

	daemonize("scard");

        while (priv->stop_thread == 0) {
		p_card_state = atomic_read(&(priv->card_existed));
		c_card_state = is_card_present(priv);
		delay = HZ / 5; /* Default 0.2 sec */
//		delay = HZ; /* Default 1 sec */

//		spin_lock(&priv->lock);
		
		///DBG_PRINT("%s: card state current %d->%d\n", priv->devname,p_card_state, c_card_state);
		
		if (p_card_state != c_card_state) {
			DBG_PRINT("%s: card state change %d->%d\n", priv->devname,
					p_card_state, c_card_state);
			/* Update the state */
			atomic_set(&(priv->card_existed), c_card_state);

			/* Detected transition of card present status */
			if (c_card_state != 0) { /* Card is here now */
				MSG_PRINT("%s: scard present detected.\n", priv->devname);

				RW_Abort = 0;

				Remove_Act_wait_count = Remove_Act = 0;
				
				Reset_mode_count = Reset_Fail_count = 0;

#ifdef READ_TIMEOUT
				Read_act = 0, Read_act_count = 0;
#endif				
				Tx_Mode = 0;
				
				memset(ATRXBuf, 0, 5);
				ATRXBuf_CNT = 0;

				atomic_set(&(priv->card_existed), 1);
				atomic_set(&(priv->state), (int)PWR_IDLE);
				
				atomic_set(&(priv->normal_mode), 0);
							
				priv->reset_cnt = 0;
				gbus_write_uint32(pGBus, SCARD_STATE_REG(priv->base_addr), 0);
				gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
						gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) & 0xffffffe0);
			} else { /* Card removed */
				MSG_PRINT("%s: scard removal detected.\n", priv->devname);

				RW_Abort = 1;

				atomic_set(&(priv->card_existed), 0);
				atomic_set(&(priv->normal_mode), 0);
				
				scard_reset(priv, 1);
				priv->rd_p_idx = priv->rd_c_idx = 0;
				priv->wr_p_idx = priv->wr_c_idx = 0;
				wake_up_interruptible(&(priv->scard_rd_wq));

				Remove_Act_wait_count = 0;
				Remove_Act = 1;

				goto out;
			}
		} else if (c_card_state != 0) {
			if (sys_freq != tangox_get_sysclock()) { /* If frequency changed, reset is needed */
				gbus_write_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr), gbus_read_uint32(pGBus, SCARD_SOFT_OUT_REG(priv->base_addr)) & ~0x40);
				gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
						gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) | (1 << 2));
				udelay(1);
				scard_reset(priv, 0); /* warm reset */
				sys_freq = tangox_get_sysclock();
				goto out;
			}
		} else if (c_card_state == 0)
			goto out;

		/* We have card here, check the state transition */
		cur_state = (enum SCARD_STATE)(gbus_read_uint32(pGBus, SCARD_STATE_REG(priv->base_addr)) & 0x7);
		if (!PERSISTENT_STATE(cur_state)) {
//			delay = HZ / 10; /* Change to 0.1 sec */
			delay = HZ / 2; /* Change to 0.5 sec */
			goto out;
		} 

		set_state = (enum SCARD_STATE)atomic_read(&(priv->state));
		if ((set_state == cur_state) || (set_state == UNKNOWN)) {
//			delay = HZ / 2;	/* State remain stable -- change to 0.5 sec */
			delay = HZ;	/* State remain stable -- change to 1 sec */
			goto out;
		} 

		DBG_PRINT("%s: expect state (%d) != current state(%d)\n", priv->devname,
				set_state, cur_state);
		if (set_state == PWR_IDLE) 
			activate_scard(priv);
		else 
			ERR_PRINT("%s: WARNING: inconsistent state %d\n", priv->devname, set_state);

out:
//		spin_unlock(&priv->lock);

#ifdef READ_TIMEOUT
		if (priv->b_mode != 0) {
			if(Read_act == 1)
			{
				#if 0
				if((Read_act_count++>=20)||(RW_Abort == 1))
				#else // 2k90722
				if((Read_act_count++>=10)||(RW_Abort == 1))
				#endif
				{
					ERR_PRINT("%s: WARNING: Read_act_count  %d\n", priv->devname, Read_act_count);
					ERR_PRINT("%s: org:priv->rd_p_idx  %d || priv->rd_c_idx  %d\n", priv->devname, priv->rd_p_idx,priv->rd_c_idx);
					Read_act_count = priv->rd_p_idx;
					priv->rd_p_idx = -1;
					wake_up_interruptible(&(priv->scard_rd_wq));
					ERR_PRINT("%s: xxx:priv->rd_p_idx  %d || priv->rd_c_idx  %d\n", priv->devname, priv->rd_p_idx,priv->rd_c_idx);
					Read_act = 2;
					Read_act_count = 0;
				}	
			}			
		}
#endif

#if 1 // 2k90214
		if(Remove_Act == 1)
		{
			if(Remove_Act_wait_count++ >= 15)
			{
				unsigned long flags;

				Remove_Act = 0;
				Remove_Act_wait_count = 0;

				MSG_PRINT("###################################################################################\n");
				MSG_PRINT("%s: scard removal detected. but Nothing scard_release.. so try to close card\n",priv->devname);
				MSG_PRINT("###################################################################################\n");
			
				spin_lock_irqsave(&priv->lock, flags);	
				gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
							gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) | ((1 << 24) | (1 << 25)));
				gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr), 
							gbus_read_uint32(pGBus, SCARD_SCARD_CTRL_REG(priv->base_addr)) & ~((1 << 24) | (1 << 25)));

				if(priv->ref_cnt > 0)
					 priv->ref_cnt--;
				priv->rd_p_idx = priv->rd_c_idx = 0;
				priv->wr_p_idx = priv->wr_c_idx = 0;
				spin_unlock_irqrestore(&priv->lock, flags);

				MSG_PRINT("###################################################################################\n");
				MSG_PRINT("%s: finish close card by scard removal detected without scard_release.. \n",priv->devname);
				MSG_PRINT("###################################################################################\n");
			}
		}
#endif

                set_current_state(TASK_UNINTERRUPTIBLE);
                schedule_timeout(delay);
        }

	return(0);
}

int __init scard_init_module(void)
{
	int status = 0;
	int i;

	extern int tangox_scard_enabled(void);
	if (tangox_scard_enabled() == 0) {
		ERR_PRINT("%s: device is disabled.\n", scard_devname);
		return(0);
	} 

	if (rd_buffer_size < 8) {
		ERR_PRINT("%s: read buffer size (%d) error, minimum 8.\n", scard_devname,
			rd_buffer_size); 
		return(-EIO);
	} else if (wr_buffer_size < 8) {
		ERR_PRINT("%s: write buffer size (%d) error, minimum 8.\n", scard_devname,
			wr_buffer_size); 
		return(-EIO);
	} 
	
	tangox_get_scard_info(&xenv_scard_5v_pin, &xenv_scard_cmd_pin, &xenv_scard_off_pin);

	/* override the default from XENV settings */
	scard_priv[0].scard_5v_pin = xenv_scard_5v_pin;
	scard_priv[0].scard_cmd_pin = xenv_scard_cmd_pin;
	scard_priv[0].scard_off_pin = xenv_scard_off_pin;

#ifdef CONFIG_TANGO3
	/* TODO: how to get information for 2nd scard controller, XENV as well? */
#endif

	/* Initialize private data structure */
	for (i = 0; i < NUM_SCARD; i++) {
		memset(&scard_priv[i], 0, sizeof(struct scard_private)); 
		spin_lock_init(&scard_priv[i].lock);

		if ((scard_priv[i].rd_buffer = kmalloc(rd_buffer_size, GFP_KERNEL)) == NULL) {
			ERR_PRINT("%s%d: out of memory for read buffer\n", scard_devname, i); 
			status = -ENOMEM;
			goto err_out;
		} else if ((scard_priv[i].wr_buffer = kmalloc(wr_buffer_size, GFP_KERNEL)) == NULL) {
			ERR_PRINT("%s%d: out of memory for write buffer\n", scard_devname, i); 
			status = -ENOMEM;
			goto err_out;
		}

		scard_priv[i].idx = i;
		scard_priv[i].base_addr = scard_ctrlers[i].base_addr;
		scard_priv[i].scard_5v_pin = scard_ctrlers[i].scard_5v_pin;
		scard_priv[i].scard_cmd_pin = scard_ctrlers[i].scard_cmd_pin;
		scard_priv[i].scard_off_pin = scard_ctrlers[i].scard_off_pin;
		scard_priv[i].scard_irq = scard_ctrlers[i].scard_irq;
		sprintf(scard_priv[i].devname, "%s%d", scard_devname, i);

		init_waitqueue_head(&(scard_priv[i].scard_rd_wq));
	}

	/* Register device, and may be allocating major# */
	status = register_chrdev(major_num, scard_devname, &scard_fops);
	if (status < 0) {
		ERR_PRINT("%s: cannot get major number\n", scard_devname); 
		goto err_out;
	} else if (major_num == 0)
		major_num = status;	/* Dynamic major# allocation */

	for (i = 0; i < NUM_SCARD; i++) {
		/* Hook up ISR */
		if (request_irq(scard_priv[i].scard_irq, scard_isr, SA_INTERRUPT, scard_priv[i].devname, &scard_priv[i]) != 0) {
			ERR_PRINT("%s: cannot register IRQ (%d)\n", scard_priv[i].devname, scard_priv[i].scard_irq);
			unregister_chrdev(major_num, scard_devname);	
			status = -EIO;
			goto err_out;
		}
	}

#ifdef CONFIG_EM86XX_SMARTCARD
	MSG_PRINT("%s: driver loaded (rd_buffer_size = %d, wr_buffer_size = %d)\n", 
		scard_devname, rd_buffer_size, wr_buffer_size);
#else
	MSG_PRINT("%s: SMP8634 driver loaded (rd_buffer_size = %d, wr_buffer_size = %d, freq = %d)\n", 
		scard_devname, rd_buffer_size, wr_buffer_size, frequency);
#endif

	sys_freq = tangox_get_sysclock();

	for (i = 0; i < NUM_SCARD; i++) {
		/* Initialize control pins */
		scard_ctrlpin_init(i);

		/* Reset the interface */
		scard_reset(&scard_priv[i], 1);

		DBG_PRINT("%s: enabled smartcard interface.\n", scard_priv[i].devname);

		scard_priv[i].stop_thread = 0;
		if (kernel_thread(scard_monitor, (void *)&scard_priv[i], CLONE_FS|CLONE_FILES) < 0)
			return(-ENODEV);
		
	}
	return(0);

err_out:
	for (i = 0; i < NUM_SCARD; i++) {
		if (scard_priv[i].rd_buffer)
			kfree(scard_priv[i].rd_buffer);
		if (scard_priv[i].wr_buffer)
			kfree(scard_priv[i].wr_buffer);
	}
	return(status);
}

void __exit scard_cleanup_module(void)
{
	extern int tangox_scard_enabled(void);
	int i;

	if (tangox_scard_enabled() == 0) 
		return;

	for (i = 0; i < NUM_SCARD; i++) {
		gbus_write_uint32(pGBus, SCARD_INTEN_REG(scard_priv[i].base_addr), 0);

		scard_priv[i].stop_thread = 1;
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(3*HZ);

		gbus_write_uint32(pGBus, SCARD_SCARD_CTRL_REG(scard_priv[i].base_addr), 0);
		gbus_write_uint32(pGBus, SCARD_STATE_REG(scard_priv[i].base_addr), 0);
		scard_ctrlpin_write(i, scard_priv[i].scard_5v_pin, 0); 
		scard_ctrlpin_write(i, scard_priv[i].scard_cmd_pin, 1); 

		unregister_chrdev(major_num, scard_devname);
		free_irq(scard_priv[i].scard_irq, &scard_priv[i]);

		if (scard_priv[i].rd_buffer != NULL)
			kfree(scard_priv[i].rd_buffer);
		if (scard_priv[i].wr_buffer != NULL)
			kfree(scard_priv[i].wr_buffer);
		
		scard_priv[i].ref_cnt = 0; // 2k90213

		MSG_PRINT("%s: driver unloaded\n", scard_devname);
	}
}

module_init(scard_init_module);
module_exit(scard_cleanup_module);

